__author__ = "David Rusk <drusk@uvic.ca>"


class ImageNotLoadedException(Exception):
    """The requested image hasn't been loaded yet."""

    def __init__(self, requested_item):
        """
        Args:
          requested_item:
            The item whose image is not loaded yet.
        """
        self.requested_item = requested_item


class NoWorkUnitException(Exception):
    """No data is available at the current time."""


class NoAvailableWorkException(Exception):
    """"No more work is available."""


class SourceNotNamedException(Exception):
    """The source has no name."""

    def __init__(self, source):
        self.source = source



# coding=utf-8
__author__ = 'Michele Bannister   git:@mtbannister'

import os
import cPickle
from collections import OrderedDict
import math

import ephem
from astropy.table import Table
from uncertainties import ufloat
import numpy
import pandas

from ossos import mpc
from ossos import orbfit
from ossos import storage
import parameters
from ossos.gui import context
import plot_lightcurve
from utils import square_fit_discovery_mag
from deluxe_table_formatter import deluxe_table_formatter


# from parameters import tno

def ossos_release_parser(table=False, data_release=parameters.RELEASE_VERSION):
    """
    extra fun as this is space-separated so using CSV parsers is not an option
    """
    names = ['cl', 'p', 'j', 'k', 'sh', 'object', 'mag', 'mag_E', 'F', 'H_sur', 'dist', 'dist_E', 'nobs',
             'time', 'av_xres', 'av_yres', 'max_x', 'max_y', 'a', 'a_E', 'e', 'e_E', 'i', 'i_E', 'node', 'node_E',
             'argperi', 'argperi_E', 'time_peri', 'time_peri_E', 'ra_dis', 'dec_dis', 'jd_dis', 'rate']#, 'eff', 'm_lim']

    if table:
        retval = storage.get_detections(release=data_release)
        # have to specify the names because the header line contains duplicate IDs, which wrecks auto-column creation
        # retval = Table.read(parameters.RELEASE_DETECTIONS[data_release], format='ascii', guess=False,
        #                    delimiter=' ', data_start=0, comment='#', names=names, header_start=None)
    else:
        retval = []
        with open(data_release, 'r') as detectionsfile:
            for line in detectionsfile.readlines()[1:]:  # first line is column definitions
                obj = TNO.from_string(line, version=parameters.RELEASE_DETECTIONS[data_release])
                retval.append(obj)

    return retval


class TNO(object):

    def __init__(self, observations):
        self.orbit = orbfit.Orbfit(observations.mpc_observations)
        try:
            # Previously discovered objects can have 'discovery-marked' lines that predate the OSSOS survey.
            discoveries = [n for n in observations.mpc_observations if (n.discovery.is_discovery and n.date > parameters.SURVEY_START)]
            # Some objects were linked to independent OSSOS-observed triples; these are also multiply 'discovery-marked'
            if len(discoveries) > 1:
                discoveries.sort(key=lambda x: x.date)
            # once sorted and restricted to within the survey timeframe, it's always the earliest such marked line.
            self.discovery = discoveries[0]
        except:
            self.discovery = False
        self.name = observations.provisional_name.split('.')[0]
        return


def ossos_discoveries(directory=parameters.REAL_KBO_AST_DIR,
                      suffix='ast',
                      no_nt_and_u=True,
                      single_object=None,
                      all_objects=False,
                      data_release=parameters.RELEASE_VERSION,
                      ):
    """
    Returns a list of objects holding orbfit.Orbfit objects with the observations in the Orbfit.observations field.
    Default is to return only the objects corresponding to the current Data Release.
    """
    retval = []
    working_context = context.get_context(directory)
    files = working_context.get_listing(suffix)

    if single_object is not None:
        files = filter(lambda name: name.startswith(single_object), files)
        print 'loading only {}'.format(files)
    elif all_objects:
        print 'loading data release {}'.format(data_release)
        # only return the objects corresponding to a particular Data Release
        data_release = ossos_release_parser(table=True, data_release=data_release)
        objects = data_release['object']
        files = filter(lambda name: name.partition(suffix)[0].rstrip('.') in objects, files)

    for filename in files:
        # keep out the not-tracked and uncharacteried.
        print filename
        if no_nt_and_u:
            if not (filename.__contains__('nt') or filename.startswith('u')):
                observations = mpc.MPCReader(directory + filename)
                obj = TNO(observations)
                retval.append(obj)
        else:  # now we want those uncharacterised ones
            observations = mpc.MPCReader(directory + filename)
            obj = TNO(observations)
            retval.append(obj)

    return retval


def ossos_release_with_metadata():
    """
    Wrap the objects from the Version Releases together with the objects instantiated from fitting their mpc lines
    """
    # discoveries = ossos_release_parser()
    discoveries = []
    observations = ossos_discoveries()
    for obj in observations:
        discov = [n for n in obj[0].mpc_observations if n.discovery.is_discovery][0]
        tno = parameters.tno()
        tno.dist = obj[1].distance
        tno.ra_discov = discov.coordinate.ra.degrees
        tno.mag = discov.mag
        tno.name = discov.provisional_name
        discoveries.append(tno)

    # for obj in discoveries:
    # observation = [n for n in observations if n.observations[-1].provisional_name == obj.name][0]
    # for obs in observation.observations:
    # if obs.discovery.is_discovery:
    #             if obj.mag is not None:
    #                 H = obj.mag + 2.5 * math.log10(1. / ((obj.dist ** 2) * ((obj.dist - 1.) ** 2)))
    #             else:
    #                 H = None
    #             obj.H = H
    #             obj.T = observation.T
    #             obj.discovery_date = obs.date
    #             obj.observations = observation

    return discoveries  # list of tno() objects


def _kbos_from_survey_sym_model_input_file(model_file):
    """
    Load a Survey Simulator model file as an array of ephem EllipticalBody objects.
    @param model_file:
    @return:
    """
    lines = storage.open_vos_or_local(model_file).read().split('\n')
    kbos = []
    for line in lines:
        if len(line) == 0 or line[0] == '#':  # skip initial column descriptors and the final blank line
            continue
        kbo = ephem.EllipticalBody()
        values = line.split()
        kbo.name = values[8]
        kbo.j = values[9]
        kbo.k = values[10]
        kbo._a = float(values[0])
        kbo._e = float(values[1])
        kbo._inc = float(values[2])
        kbo._Om = float(values[3])
        kbo._om = float(values[4])
        kbo._M = float(values[5])
        kbo._H = float(values[6])
        epoch = ephem.date(2453157.50000 - ephem.julian_date(0))
        kbo._epoch_M = epoch
        kbo._epoch = epoch
        kbos.append(kbo)
    return kbos


def kbos_from_survey_sym_model_input_file(infile):
    # names = ['a', 'e', 'i', 'node', 'peri', 'M', 'H', 'distance', 'orbtype', 'j', 'k']

    # 3 components (hot [h], cold kernel [c] and cold stirred [f])
    names = ['a', 'e', 'i', 'node', 'peri', 'M', 'H', 'distance', 'component', 'orbtype']

    kbos = pandas.read_table(infile, names=names, delim_whitespace=True, comment='#')

    return kbos


def synthetic_model_kbos(at_date=parameters.NEWMOONS[parameters.DISCOVERY_NEW_MOON], maglimit=24.5, kbotype=False,
                         arrays=False):
    # # build a list of Synthetic KBOs
    kbos = []
    print "LOADING SYNTHETIC MODEL KBOS FROM: {}".format(parameters.L7MODEL)
    pastpath = parameters.L7_HOME + str(at_date).split()[0].replace('/', '-') + '_' + str(maglimit) + (
        kbotype or '') + '.dat'
    print(pastpath)

    # Load the previously cached orbits, quick.
    if os.path.exists(pastpath) and arrays:
        with open(pastpath) as infile:
            ra, dist, hlat, Hmag = cPickle.load(infile)
        print('{} synthetic model kbos brighter than {} at {} in L7 model'.format(len(ra), maglimit, at_date))
        return ra, dist, hlat, Hmag

    model_kbos = _kbos_from_survey_sym_model_input_file(parameters.L7MODEL)

    counter = 0
    ra = []
    dist = []
    hlat = []
    Hmag = []

    for kbo in model_kbos:
        if kbotype and (kbo.name == kbotype) and (kbo.j == '3' and kbo.k == '2'):  # keeping 3:2 resonators
            date = ephem.date(at_date)
            kbo.compute(date)
            counter += 1
            # ## only keep objects that are brighter than limit
            if kbo.mag < maglimit:
                kbos.append(kbo)
                ra.append(float(kbo.ra))
                dist.append(float(kbo.sun_distance))
                hlat.append(float(kbo.hlat))
                Hmag.append(float(kbo._H))

    print '{} synthetic model kbos brighter than {} at {} retained from {} in L7 model'.format(len(kbos), maglimit,
                                                                                               at_date, counter)
    if not arrays:
        return kbos
    else:
        with open(pastpath, 'w') as outfile:
            cPickle.dump((ra, dist, hlat, Hmag), outfile)
        return ra, dist, hlat, Hmag


def output_discoveries_for_animation():
    discoveries = ossos_release_with_metadata()
    with open('/Users/michele/Desktop/OSSOSdiscoveries.txt', 'w') as outfile:
        outfile.write('name a e i node peri MA epoch H date_of_discovery\n')
        for obj in discoveries:
            # a e i node peri MA epoch H date_of_discovery
            outfile.write(
                '{:>10s} {:8.2f} {:8.2f} {:8.2f} {:8.2f} {:8.2f} {:8.2f} {:8.2f} {:8.2f} {:>10s}\n'.format(
                    obj.name, obj.a, obj.e, obj.i, obj.node, obj.argp, obj.M, obj.T, obj.H, obj.discovery_date))


def round_sig_error(num, uncert):
    """
    Return a string of the number and its uncertainty to the right sig figs via uncertainty's print methods.
    The uncertainty determines the sig fig rounding of the number.
    https://pythonhosted.org/uncertainties/user_guide.html
    """
    u = ufloat(num, uncert)
    return '{:.1uLS}'.format(u)


def linesep(name, distinguish=None):
    names = {'N': "Objects in resonance with Neptune",
             'U': "Objects in resonance with Uranus",  # don't have one yet but Kat does check
             'i': 'Inner classical belt',
             'm': 'Main classical belt',
             'o': 'Outer classical belt',
             'd': 'Detached classical belt',
             'x': 'Scattering disk',
             'c': 'Centaurs',
             }
    if distinguish:
        if distinguish == 'sca':
            name = 'x'
        elif distinguish == 'det':
            name = 'd'
        elif distinguish == 'cen':
            name = 'c'

    midline = r"\cutinhead{" + \
              "{}".format(names[name]) + \
              "} \n"

    return midline


def extract_mpc_designations(initial_id='o3'):
    # Scrape the index file for MPC designations, if they exist
    idx = {}
    with open(parameters.IDX) as infile:
        lines = infile.readlines()
        for line in lines:
            key = line.split(' ')[0].split('\t')[0]
            if key.startswith(initial_id):
                ls = line[14:].split(' ')
                if len(ls) == 3:
                    ls = ls[1:]
                mpcstr = "{} {}".format(ls[0], ls[1][0:2]) + r"$_{" + "{}".format(ls[1][2:].rstrip('\r\n')) + r"}$"
                idx[key] = mpcstr

    return idx


def extract_highest_designation(initial_id):
    # Scrape the index file for MPC designations, if they exist
    idx = {}
    with open(parameters.IDX) as infile:
        lines = infile.readlines()
        for line in lines:
            designations = line.split(' ')
            if initial_id in designations:
                # Highest internal OSSOS one will be leftmost. FIXME: some logic for getting MPC designations, to right
                return designations[0].split('\t')[0]


def create_table(tnos, outfile, initial_id='o3'):
    # sort order gives Classical, Detached, Resonant, Scattered
    # Sort by discovery mag within each classification.
    tnos.sort(['cl', 'p', 'j', 'k', 'mag'])
    # FIXME: add a catch if object doesn't yet have a MPC designation
    idx = extract_mpc_designations(initial_id=initial_id)

    with deluxe_table_formatter(outfile) as ofile:
        for i, r in enumerate(tnos):
            # a labelled line separator heading for each new object classification type
            if r['p'] != tnos[i - 1]['p']:
                if r['p'] == 'x':  # 'x' doesn't give enough info to set scattered or detached: go check
                    ofile.write(linesep(r['p'], distinguish=r['cl']))
                else:
                    ofile.write(linesep(r['p']))

            # Mag uncertainty is over the whole light curve. Uncharacterised objects might not have enough valid points.
            sigma_mag = plot_lightcurve.stddev_phot(r['object'])
            if not numpy.isnan(sigma_mag):
                sigma_mag = '{:2.2f}'.format(sigma_mag)
            else:
                sigma_mag = r'--'

            # Individual object discovery efficiency from the function based on its mag + motion rate at discovery
            eff_at_discovery = square_fit_discovery_mag(r['object'], r['mag'], r['rate'])

            # mag ± dmag, std dev of all clean photometry, efficiency function at that discovery mag
            # m ± dm sigma_m eff_discov RA Dec a ± da e ± de i ± di r ± dr H ± dH j k MPC obj status
            out = "{} & {} & {:2.2f} & {:3.3f} & {} & {} & {} & {} & {} & {} & {} & {} & ".format(
                round_sig_error(r['mag'], r['mag_E']),
                sigma_mag,
                eff_at_discovery,
                math.degrees(ephem.degrees(ephem.hours(str(r['ra_dis'])))),
                r['dec_dis'],
                round_sig_error(r['a'], r['a_E']),
                round_sig_error(r['e'], r['e_E']),
                round_sig_error(r['i'], r['i_E']),
                round_sig_error(r['dist'], r['dist_E']),
                r['H_sur'],
                idx[r['object']],  # MPC designation is already formatted.
                r['object']
            )
            # output the orbit classification with nice formatting, or leave a gap if unclassified (e.g. a classical)
            if r['j'] != -1:
                out += "{}:{} ".format(r['j'], r['k'])  # resonant object: give the resonance
            else:
                out += "    "  # it's a classical or scattered object
            if r['sh'] != 'S':
                out += "{} {} \n".format(r['sh'], r'\\')
            else:
                out += "  {} \n".format(r'\\')
            ofile.write(out)


def release_to_latex(outfile):
    tnos = ossos_release_parser(table=True)
    characterised = tnos[numpy.array([name.startswith("o") for name in tnos['object']])]
    create_table(characterised, outfile, initial_id='o3')

    # uncharacterised = tnos[numpy.array([name.startswith("u") for name in tnos['object']])]
    # create_table(uncharacterised, 'u_' + outfile)

    return


def parse_subaru_radec(line):
    d = line.split()
    pointing_name = line.split('=')[0]
    ra = d[1].split('=')[1]
    dec = d[2].split('=')[1]
    if len(ra.split('.')[0]) == 5:  # LACK OF SEPARATORS ARGH
        ra = '0' + ra
    if len(dec.split('.')[0]) == 5:
        dec = '0' + dec
    ra = "{}:{}:{}".format(ra[0:2], ra[2:4], ra[4:])
    dec = "{}:{}:{}".format(dec[0:2], dec[2:4], dec[4:])
    return pointing_name, ra, dec


def parse_subaru_mags():
    tnos = ossos_discoveries(parameters.REAL_KBO_AST_DIR, suffix='ast')
    index = OrderedDict()
    with open('/Users/michele/Desktop/Col3N.txt', 'r') as infile:
        lines = infile.readlines()
        for line in lines:
            pointing, ra, dec = parse_subaru_radec(line)
            objs = [t for t in pointing.split('_') if t not in ['only', 'alternate', 'bonus']]
            mags = []
            regular = []
            for obj in objs:
                mag = [t.discovery.mag for t in tnos if t.name.endswith(obj)][0]
                mags.append(mag)
                if len([t for t in parameters.COLOSSOS if t.endswith(obj)]) == 0:
                    regular.append(obj)
            index[pointing] = (ra, dec, mags, regular)
            print pointing, index[pointing]

    with open('/Users/michele/Desktop/Col3N_201509121000.txt', 'w') as outfile:
        outfile.write('{: <25} {: <15} {: <15} {: <30} {: <10}\n'.format('Pointing',
                                                                         'RA (hh:mm:ss)',
                                                                         'Dec (dd:mm:ss)',
                                                                         'magnitudes at discovery (m_r)',
                                                                         'regular OSSOS'))
        for key, val in index.items():
            outfile.write('{: <25} {: <15} {: <15} {: <30} {: <10}\n'.format(key,
                                                                             val[0],
                                                                             val[1],
                                                                             ', '.join([str(f) for f in val[2]]),
                                                                             ', '.join([f for f in val[3]])
                                                                             ))


def block_table_pprint():
    with open('block_table.tex', 'w') as outfile:
        for name, coords in parameters.BLOCKS.items():
            print name, coords
            ra = ephem.hours(coords['RA'])
            dec = ephem.degrees(coords['DEC'])
            eq = ephem.Equatorial(ra, dec)
            ec = ephem.Ecliptic(eq)
            outfile.write("{} & {:2.1f} & {:2.1f} & {:2.1f} & {:2.1f} {} \n".format(
                name[2:],
                math.degrees(ephem.degrees(ra)),
                math.degrees(dec),
                math.degrees(ec.lat),
                math.degrees(ec.lon),
                r"\\"))
    return


if __name__ == '__main__':
    # ossos_release_parser(table=True)
    release_to_latex('v{}'.format(parameters.RELEASE_VERSION) + '_table.tex')
    # parse_subaru_mags()
    # block_table_pprint()



__author__ = "David Rusk <drusk@uvic.ca>"


class FocusCalculator(object):
    def convert_source_location(self, source_reading, reference_reading):
        """
        Converts the source (x, y) location from reading into the coordinate
        system of reference_reading.
        """
        offset_x, offset_y = reference_reading.get_coordinate_offset(source_reading)
        focus = source_reading.x + offset_x, source_reading.y + offset_y
        return focus


class SingletFocusCalculator(FocusCalculator):
    """
    Calculates focal points for display as single images.  The focal point
    will be the location of the source in the middle observation.
    """

    def __init__(self, source):
        self.source = source

    def calculate_focus(self, reading):
        """
        Determines what the focal point of the downloaded image should be.

        Returns:
          focal_point: (x, y)
            The location of the source in the middle observation, in the
            coordinate system of the current source reading.
        """
        middle_index = len(self.source.get_readings()) // 2
        middle_reading = self.source.get_reading(middle_index)
        return self.convert_source_location(middle_reading, reading)


class TripletFocusCalculator(FocusCalculator):
    """
    Calculates the focal points for displaying triplets.
    """

    def __init__(self, source):
        self.source = source

    def calculate_focus(self, reading, frame_index):
        center_target = self.source.get_reading(frame_index)
        return self.convert_source_location(center_target, reading)



import traceback

__author__ = "David Rusk <drusk@uvic.ca>"

import errno
import os

import requests

from ossos.gui import logger


class DownloadErrorHandler(object):
    def __init__(self, app):
        self.app = app

        self._failed_downloads = []

    def handle_error(self, error, download_request):
        """
        Checks what error occured and looks for an appropriate solution.

        Args:
          error: Exception
            The error that has occured.
          download_request:
            The request which resulted in the error.
        """
        if hasattr(error, "errno") and error.errno == errno.EACCES:
            self.handle_certificate_problem(str(error))
        else:
            self.handle_general_download_error(str(error), download_request)

    def handle_certificate_problem(self, error_message):
        logger.warning("Problem with CADC certificate")

        self.app.get_model().stop_loading_images()

        view = self.app.get_view()
        view.hide_image_loading_dialog()
        view.show_certificate_dialog(self, error_message)

    def refresh_certificate(self, username, password):
        download_certificate(username, password)

        logger.info("Downloaded new CADC certificate")

        model = self.app.get_model()
        model.refresh_vos_client()

        self.app.get_view().show_image_loading_dialog()
        model.start_loading_images()

    def handle_general_download_error(self, error_message, download_request):
        logger.warning("A problem occurred while downloading: %s" % error_message)
        logger.error("-" * 60)
        logger.error(traceback.format_exc())
        logger.error("-" * 60)

        self._failed_downloads.append(download_request)
        self.app.get_view().show_retry_download_dialog(self, error_message)

    def retry_downloads(self):
        model = self.app.get_model()
        for download_request in self._failed_downloads:
            model.submit_download_request(download_request)


def download_certificate(username, password):
    url = "http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/cred/proxyCert?daysValid=7"
    response = requests.get(url, auth=(username, password))

    certfile = os.path.join(os.getenv("HOME"), ".ssl/cadcproxy.pem")
    with open(certfile, "wb") as filehandle:
        filehandle.write(response.content)



__author__ = "David Rusk <drusk@uvic.ca>"

import collections
import threading

from .. import storage
from .. import auth
from . import tasks


CANDS = "CANDS"
REALS = "REALS"

# Constants for local progress manager
DONE_SUFFIX = ".DONE"
LOCK_SUFFIX = ".LOCK"
PART_SUFFIX = ".PART"

# Constants for VOSpace progress manager
DONE_PROPERTY = "done"
PROCESSED_INDICES_PROPERTY = "processed_indices"
LOCK_PROPERTY = "lock_holder"

# TODO: just make them both "," for consistency
INDEX_SEP = "\n"
VO_INDEX_SEP = ","


def requires_lock(function):
    """
    Decorator to check if the user owns the required lock.
    The first argument must be the filename.
    """

    def new_lock_requiring_function(self, filename, *args, **kwargs):
        if self.owns_lock(filename):
            return function(self, filename, *args, **kwargs)
        else:
            raise RequiresLockException()

    return new_lock_requiring_function


class FileLockedException(Exception):
    """Indicates someone already has a lock on the requested file."""

    def __init__(self, filename, locker):
        self.filename = filename
        self.locker = locker

        super(FileLockedException, self).__init__(
            "%s is locked by %s" % (filename, locker))


class RequiresLockException(Exception):
    def __init__(self):
        super(RequiresLockException, self).__init__(
            "Operation requires a lock on the file.")


class AbstractProgressManager(object):
    """
    Manages persistence of progress made processing files in a directory.
    """

    def __init__(self, working_context, userid=None):
        self.working_context = working_context
        if userid is None:
            self.userid = auth.get_cadc_username()
        else:
            self.userid = userid

    def get_done(self, task):
        """
        Gets a list of the files which have already been completely
        processed for a task.

        Args:
          task: str
            The task to find files for.

        Returns:
          done_list: list(str)
            The list of filenames that have been completely processed.
        """
        raise NotImplementedError()

    def is_done(self, filename):
        """
        Checks if a file has been completely processed.

        Args:
          filename: str
            A file in the working directory to check for completion.
            No lock is required for this operation.

        Returns:
          is_done: bool
            True if all items in the file have been processed,
            False otherwise.
        """
        raise NotImplementedError()

    def get_processed_indices(self, filename):
        """
        Retrieve indices of items that have been processed in a file.

        Args:
          filename: str
            A file in the working directory to check the progress on.
            No lock is required for this operation.

        Returns:
          processed_indices: list(int)
            The indices of items in the specified field that have been
            processed already.  Returns an empty list if no items have
            been processed.
        """
        raise NotImplementedError()

    @requires_lock
    def record_done(self, filename):
        """
        Records a file as being completely processed.

        Args:
          filename: str
            A file in the working directory to mark as processed.
            The caller MUST own the lock on this file.

        Returns: void

        NOTE: Does not remove the caller's lock on the file.
        """
        self._record_done(filename)

    def _record_done(self, filename):
        raise NotImplementedError()

    @requires_lock
    def record_index(self, filename, index):
        """
        Records that an item at a specific index within a file has
        been processed.

        Args:
          filename: str
            The file in the working directory which contains the index
            which is being marked as processed.
            The caller MUST own the lock on this file.
          index: int
            The 0-based index of the item that has been processed.

        Returns: void
        """
        self._record_index(filename, index)

    def _record_index(self, filename, index):
        raise NotImplementedError()

    def lock(self, filename):
        raise NotImplementedError()

    def unlock(self, filename, async=False):
        """
        Unlocks the file.

        Args:
          filename: str
            The file to be unlocked.
          async: bool
            If set to True, tries to unlock the file asynchronously, if
            possible.  This allows processing to continue if it is not
            critical that the lock be released before continuing.

        Returns: void
        """
        raise NotImplementedError()

    def clean(self, suffixes=None):
        raise NotImplementedError()

    def owns_lock(self, filename):
        raise NotImplementedError()


class VOSpaceProgressManager(AbstractProgressManager):
    def __init__(self, working_context, userid=None, track_partial_progress=False):
        """
        By default partial results are not tracked when working in VOSpace.
        get_processed_indices returns an empty list and record_index is a
        no-op.
        """
        super(VOSpaceProgressManager, self).__init__(working_context, userid=userid)

        self.track_partial_results = track_partial_progress

    def get_done(self, task):
        return [filename for filename in self.working_context.get_listing(task)
                if self.is_done(filename)]

    def is_done(self, filename):
        return storage.has_property(self._get_uri(filename), DONE_PROPERTY)

    def get_processed_indices(self, filename):
        if not self.track_partial_results:
            return []

        uri = self._get_uri(filename)
        if not storage.has_property(uri, PROCESSED_INDICES_PROPERTY):
            return []

        raw_property = storage.get_property(uri, PROCESSED_INDICES_PROPERTY)
        return map(int, raw_property.split(VO_INDEX_SEP))

    def _record_done(self, filename):
        storage.set_property(self._get_uri(filename), DONE_PROPERTY,
                             self.userid)

    def _record_index(self, filename, index):
        if not self.track_partial_results:
            return

        processed_indices = self.get_processed_indices(filename)
        processed_indices.append(index)
        processed_indices = map(str, processed_indices)

        storage.set_property(self._get_uri(filename),
                             PROCESSED_INDICES_PROPERTY,
                             VO_INDEX_SEP.join(processed_indices))

    def lock(self, filename):
        uri = self._get_uri(filename)

        lock_holder = storage.get_property(uri, LOCK_PROPERTY)
        if lock_holder is None:
            storage.set_property(uri, LOCK_PROPERTY, self.userid)
        elif lock_holder == self.userid:
            # We already had the lock
            pass
        else:
            raise FileLockedException(filename, lock_holder)

    def unlock(self, filename, async=False):
        if async:
            threading.Thread(target=self._do_unlock, args=(filename, )).start()
        else:
            self._do_unlock(filename)

    def _do_unlock(self, filename):
        uri = self._get_uri(filename)

        lock_holder = storage.get_property(uri, LOCK_PROPERTY)
        if lock_holder is None:
            # The file isn't actually locked.  Probably already cleaned up.
            pass
        elif lock_holder == self.userid:
            # It was us who locked it
            storage.set_property(uri, LOCK_PROPERTY, None)
        else:
            # Can't remove someone else's lock!
            raise FileLockedException(filename, lock_holder)

    def clean(self, suffixes=None):
        pass

    def owns_lock(self, filename):
        lock_holder = storage.get_property(self._get_uri(filename),
                                           LOCK_PROPERTY)
        return lock_holder == self.userid

    def _get_uri(self, filename):
        return self.working_context.get_full_path(filename)


class LocalProgressManager(AbstractProgressManager):
    """
    Persists progress locally to disk.
    """

    def __init__(self, working_context, userid=None):
        super(LocalProgressManager, self).__init__(working_context, userid=userid)

    def get_done(self, task):
        listing = self.working_context.get_listing(self._get_done_suffix(task))
        return [done_file[:-len(DONE_SUFFIX)] for done_file in listing]

    def is_done(self, filename):
        return self.working_context.exists(filename + DONE_SUFFIX)

    def get_processed_indices(self, filename):
        partfile = filename + PART_SUFFIX
        donefile = filename + DONE_SUFFIX

        file_with_records = None
        if self.working_context.exists(donefile):
            file_with_records = donefile
        elif self.working_context.exists(partfile):
            file_with_records = partfile

        if file_with_records is None:
            return []

        filehandle = self.working_context.open(file_with_records)
        indices = filehandle.read().rstrip(INDEX_SEP).split(INDEX_SEP)
        filehandle.close()

        return map(int, indices)

    def _record_done(self, filename):
        partfile = filename + PART_SUFFIX
        donefile = filename + DONE_SUFFIX

        if self.working_context.exists(partfile):
            # By just renaming we keep the history of processed indices
            # available.
            self.working_context.rename(partfile, donefile)
        else:
            # Create a new blank file
            self.working_context.open(donefile).close()

    def _record_index(self, filename, index):
        filehandle = self.working_context.open(filename + PART_SUFFIX)
        filehandle.write(str(index) + INDEX_SEP)
        filehandle.close()

    def lock(self, filename):
        if self.owns_lock(filename):
            return

        lockfile = filename + LOCK_SUFFIX
        if self.working_context.exists(lockfile):
            locker = self.working_context.open(lockfile).read()
            raise FileLockedException(filename, locker)
        else:
            filehandle = self.working_context.open(lockfile)
            filehandle.write(self.userid)
            filehandle.close()

    def unlock(self, filename, async=False):
        # NOTE: locally this is fast so we don't both doing it asynchronously
        lockfile = filename + LOCK_SUFFIX

        if not self.working_context.exists(lockfile):
            # The lock file was probably already cleaned up by record_done
            return

        filehandle = self.working_context.open(lockfile)
        locker = filehandle.read()
        filehandle.close()

        if locker == self.userid:
            # It was us who locked it
            self.working_context.remove(lockfile)
        else:
            # Can't remove someone else's lock!
            raise FileLockedException(filename, locker)

    def clean(self, suffixes=None):
        """
        Remove all persistence-related files from the directory.
        """
        if suffixes is None:
            suffixes = [DONE_SUFFIX, LOCK_SUFFIX, PART_SUFFIX]

        for suffix in suffixes:
            listing = self.working_context.get_listing(suffix)
            for filename in listing:
                self.working_context.remove(filename)

    def owns_lock(self, filename):
        lockfile = filename + LOCK_SUFFIX

        if self.working_context.exists(lockfile):
            filehandle = self.working_context.open(lockfile)
            lock_holder = filehandle.read()
            filehandle.close()
            return self.userid == lock_holder
        else:
            # No lock file, so we can't have a lock
            return False

    def _atomic_create(self, filename):
        """
        NOTE: this is not currently being used.

        Tries to create the specified file.  Throws an OSError if it already
        exists.
        """
        # fd = os.open(full_path, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
        # return os.fdopen(fd, "wb")
        pass

    def _get_done_suffix(self, task):
        return tasks.get_suffix(task) + DONE_SUFFIX


class InMemoryProgressManager(AbstractProgressManager):
    """
    An implementation of the ProgressManager interface which stores all
    information in memory.  It is mainly intended for convenience in
    testing.
    """

    def __init__(self, working_context, userid=None):
        super(InMemoryProgressManager, self).__init__(working_context, userid=userid)
        self.done = set()
        self.owned_locks = set()
        self.external_locks = set()
        self.processed_indices = collections.defaultdict(list)

    def get_done(self, task):
        listing = self.working_context.get_listing(self._get_done_suffix(task))
        return [done_file[:-len(DONE_SUFFIX)] for done_file in listing]

    def is_done(self, filename):
        return filename in self.done

    def get_processed_indices(self, filename):
        return self.processed_indices[filename]

    def _record_done(self, filename):
        self.done.add(filename)

    def _record_index(self, filename, index):
        self.processed_indices[filename].append(index)

    def lock(self, filename):
        if filename in self.external_locks:
            raise FileLockedException(filename, "x")

        self.owned_locks.add(filename)

    def unlock(self, filename, async=False):
        if filename in self.external_locks:
            raise FileLockedException(filename, "x")

        if filename in self.owned_locks:
            self.owned_locks.remove(filename)

    def owns_lock(self, filename):
        return filename in self.owned_locks

    def add_external_lock(self, filename):
        self.external_locks.add(filename)

    def _get_done_suffix(self, task):
        return tasks.get_suffix(task) + DONE_SUFFIX



__author__ = "David Rusk <drusk@uvic.ca>"
import json
import os

DEFAULT_CONFIG_FILE = "config.json"

_configs = {}


def read(keypath, configfile=None):
    """
    Reads a value from the configuration file.

    Args:
      keypath: str
        Specifies the key for which the value is desired.  It can be a
        hierarchical path.  Example: "section1.subsection.key1"
      configfile: str
        Path to the config file to read.  Defaults to None, in which case
        the application's default config file is used.

    Returns:
      value from configuration file
    """
    if configfile in _configs:
        appconfig = _configs[configfile]
    else:
        appconfig = AppConfig(configfile=configfile)
        _configs[configfile] = appconfig

    return appconfig.read(keypath)


class AppConfig(object):
    """
    Provides programmatic access to contents of the application
    configuration file.
    """

    def __init__(self, configfile=None):
        if configfile is None:
            configfile = self._get_default_config_file_path()

        with open(configfile, "rb") as filehandle:
            self._data = json.loads(filehandle.read())

    def _get_default_config_file_path(self):
        return os.path.join(os.path.dirname(__file__), DEFAULT_CONFIG_FILE)

    def read(self, keypath):
        curr_data = self._data
        for key in keypath.split("."):
            curr_data = curr_data[key]
        return curr_data



__author__ = "David Rusk <drusk@uvic.ca>"

import wx
import wx.lib.inspection

from .. import logger
import dialogs
from errorhandling import CertificateDialog, RetryDownloadDialog
from imageview import ImageViewManager
from keybinds import KeybindManager
from loading import WaitingGaugeDialog
from mainframe import MainFrame
from menu import Menu
from validation import AcceptSourceDialog, RejectSourceDialog, OffsetSourceDialog


def guithread(function):
    """
    Decorator to make sure the function is called from wx's main GUI
    thread.  This is needed when trying to trigger UI updates from a thread
    other than the main GUI thread (such as some asynchronous data loading
    thread).

    I learned about doing this from:
    wxPython 2.8 Application Development Cookbook, Chapter 11
    """

    def new_guithread_function(*args, **kwargs):
        if wx.Thread_IsMain():
            return function(*args, **kwargs)
        else:
            wx.CallAfter(function, *args, **kwargs)

    return new_guithread_function


class ApplicationView(object):
    """
    Provides the view's external interface.
    """

    def __init__(self, controller_factory, track_mode=False, debug=False, mark_using_pixels=False, zoom=1):
        logger.debug("Creating Application View")
        self.controller = controller_factory.create_controller(self)

        self.wx_app = wx.App(False)
        logger.debug("App built")
        self.debug = debug
        self.mark_using_pixels = mark_using_pixels

        self.mainframe = MainFrame(self.controller, track_mode=track_mode)
        self.image_view_manager = ImageViewManager(self.mainframe, zoom)
        self.menu = Menu(self.mainframe, self.controller)
        self.keybind_manager = KeybindManager(self.mainframe, self.controller)

        self.loading_dialog = WaitingGaugeDialog(self.mainframe,
                                                 "Image loading...")

        # Handle user clicking on the window's "x" button
        self.mainframe.Bind(wx.EVT_CLOSE, self._on_close_window)

        self.accept_source_dialog = None
        self.reject_source_dialog = None
        self.certificate_dialog = None
        self.retry_downloads_dialog = None

        # TODO refactor
        self.register_xy_changed_event_handler(self.controller.on_reposition_source)
        logger.debug("Done.")

    def _on_close_window(self, event):
        self.close()

    @property
    def image_viewer(self):
        return self.image_view_manager.image_viewer

    @guithread
    def show(self):
        self.mainframe.Show()

        if self.debug:
            logger.info("Launching view in debug mode.")
            wx.lib.inspection.InspectionTool().Show()

        self.wx_app.MainLoop()

    @guithread
    def display(self, cutout, use_pixel_coords=False):
        self.image_viewer.display(cutout, use_pixel_coords)

    @guithread
    def place_marker(self, cutout, x, y, radius=10, colour='r'):
        self.image_viewer.place_marker(cutout, x, y, radius, colour)

    @property
    def ds9(self):
        return self.image_viewer.ds9

    @guithread
    def align(self, cutout, reading, source):
        self.image_viewer.align(cutout, reading, source)

    @guithread
    def clear(self):
        self.image_viewer.clear()

    @guithread
    def frame(self, frame):
        return self.image_viewer.ds9.set('frame {}'.format(frame))

    @guithread
    def refresh_markers(self):
        self.image_viewer.refresh_markers()

    @guithread
    def draw_uncertainty_ellipse(self, cutout):
        self.image_viewer.draw_uncertainty_ellipse(cutout)

    @guithread
    def mark_apertures(self, cutout, pixel=False):
        self.image_viewer.mark_apertures(cutout, pixel)

    @guithread
    def reset_colormap(self):
        self.image_viewer.reset_colormap()

    @guithread
    def toggle_reticule(self):
        self.image_viewer.toggle_reticule()

    @guithread
    def register_xy_changed_event_handler(self, handler):
        self.image_viewer.register_xy_changed_event_handler(handler)

    @guithread
    def show_image_loading_dialog(self):
        if not self.loading_dialog.IsShown():
            self.loading_dialog.CenterOnParent()
            self.loading_dialog.Show()

    @guithread
    def hide_image_loading_dialog(self):
        if self.loading_dialog.IsShown():
            self.loading_dialog.Hide()

    @guithread
    def use_singlets(self):
        self.image_view_manager.use_singlets()

    @guithread
    def use_triplets(self):
        self.image_view_manager.use_triplets()

    @guithread
    def update_displayed_data(self, reading_data, header_data_list):
        self.mainframe.update_displayed_data(reading_data, header_data_list)

    @guithread
    def set_observation_status(self, current_obs, total_obs):
        self.mainframe.set_observation_status(current_obs, total_obs)

    @guithread
    def enable_source_validation(self):
        self.mainframe.enable_validation()

    @guithread
    def disable_source_validation(self):
        self.mainframe.disable_validation()

    def is_source_validation_enabled(self):
        return self.mainframe.is_source_validation_enabled()

    @guithread
    def disable_sync_menu(self):
        self.menu.disable_sync()

    @guithread
    def show_certificate_dialog(self, handler, error_message):
        if not self.certificate_dialog:
            self.certificate_dialog = CertificateDialog(self.mainframe,
                                                        handler, error_message)
            self.certificate_dialog.ShowModal()

    @guithread
    def show_retry_download_dialog(self, handler, error_message):
        # Only allow one dialog to be shown at a time
        if not self.retry_downloads_dialog:
            self.retry_downloads_dialog = RetryDownloadDialog(
                self.mainframe, handler, error_message)
            self.retry_downloads_dialog.Show()

    @guithread
    def show_accept_source_dialog(self, provisional_name,
                                  date_of_obs,
                                  ra,
                                  dec,
                                  obs_mag,
                                  obs_mag_err,
                                  band,
                                  note1_choices=None,
                                  note2_choices=None,
                                  note1_default=None,
                                  note2_default=None,
                                  default_observatory_code="",
                                  default_comment="",
                                  phot_failure=False,
                                  pixel_x=None,
                                  pixel_y=None
    ):
        self.accept_source_dialog = AcceptSourceDialog(
            self.mainframe, self.controller,
            provisional_name,
            date_of_obs,
            ra,
            dec,
            obs_mag,
            obs_mag_err,
            band,
            note1_choices=note1_choices,
            note2_choices=note2_choices,
            note1_default=note1_default,
            note2_default=note2_default,
            default_observatory_code=default_observatory_code,
            default_comment=default_comment,
            phot_failure=phot_failure,
            pixel_x=pixel_x,
            pixel_y=pixel_y)
        self.accept_source_dialog.ShowModal()

    @guithread
    def close_accept_source_dialog(self):
        if self.accept_source_dialog is not None:
            self.accept_source_dialog.Close()
            self.accept_source_dialog = None

    @guithread
    def show_reject_source_dialog(self):
        self.reject_source_dialog = RejectSourceDialog(
            self.mainframe, self.controller)
        self.reject_source_dialog.ShowModal()

    @guithread
    def close_reject_source_dialog(self):
        if self.reject_source_dialog is not None:
            self.reject_source_dialog.Close()
            self.reject_source_dialog = None

    @guithread
    def show_offset_source_dialog(self, cen_coords, pix_coords):
        self.offset_source_dialog = OffsetSourceDialog(
            self.mainframe, self.controller, cen_coords, pix_coords)
        self.offset_source_dialog.ShowModal()

    @guithread
    def close_offset_source_dialog(self):
        if self.offset_source_dialog is not None:
            self.offset_source_dialog.Close()
            self.offset_source_dialog = None

    @guithread
    def show_keymappings(self):
        dialogs.show_keymappings_dialog(self.mainframe, self.keybind_manager)

    @guithread
    def show_empty_workload_dialog(self, directory):
        dialogs.show_empty_workload_dialog(self.mainframe, directory)

    @guithread
    def all_processed_should_exit_prompt(self):
        return dialogs.should_exit_prompt(self.mainframe)

    @guithread
    def set_autoplay(self, autoplay_enabled):
        self.menu.set_autoplay(autoplay_enabled)

    def as_widget(self):
        return self.mainframe

    def close(self):
        self.mainframe.Destroy()



__author__ = "David Rusk <drusk@uvic.ca>"

import copy

from astropy.io.fits.hdu.hdulist import HDUList
from mock import Mock, MagicMock


class CopyingMock(Mock):
    """
    Useful when you want to make assertions about mutable call parameters.
    """

    def __call__(self, *args, **kwargs):
        args = copy.deepcopy(args)
        kwargs = copy.deepcopy(kwargs)
        return super(CopyingMock, self).__call__(*args, **kwargs)


class Dummy(object):
    """
    Create a dummy object that can be used as a substitute in test code.
    When any method gets called on it, it will print out the fact that the
    method was called.
    """

    def __init__(self, dummyname=None):
        if dummyname is not None:
            self._dummyname = dummyname
        else:
            self._dummyname = "Dummy"

    def __getattribute__(self, name):
        if name.startswith("_"):
            # This attribute is private to the Dummy class, access it
            # normally.
            attr = object.__getattribute__(self, name)
            if hasattr(attr, "__call__"):
                return attr()
            else:
                return attr

        def dummy_function(*args, **kwargs):
            print "Called %s on %s" % (name, self._dummyname)

        return dummy_function


def mock_hdulist():
    return MagicMock(spec=HDUList)



__author__ = "David Rusk <drusk@uvic.ca>"

import threading

from ossos.gui import config


class AutoplayManager(object):
    def __init__(self, model):
        self.model = model

        self.interval = config.read("DISPLAY.AUTOPLAY_INTERVAL")
        self._task_thread = None
        self._running = False

    def start_autoplay(self):
        self._task_thread = AutoplayThread(self.model, self.interval)
        self._task_thread.start()
        self._running = True

    def stop_autoplay(self):
        if self._running:
            self._task_thread.stop()
            self._running = False

    def is_running(self):
        return self._running


class TaskThread(threading.Thread):
    """
    A thread that executes a task at specified intervals.
    """

    def __init__(self, interval):
        """
        Args:
          interval: float
            The interval between task executions in seconds.
        """
        super(TaskThread, self).__init__()

        self._finished = threading.Event()
        self._interval = interval

    def stop(self):
        self._finished.set()

    def run(self):
        while not self._finished.is_set():
            self.do_task()

            # Sleep for interval or until stopped
            self._finished.wait(self._interval)

    def do_task(self):
        """
        The task done by this thread.  Provide functionality by overriding
        in subclasses.
        """
        pass


class AutoplayThread(TaskThread):
    def __init__(self, model, interval):
        super(AutoplayThread, self).__init__(interval)
        self.model = model

    def do_task(self):
        self.model.next_obs()




__author__ = 'Michele Bannister   git:@mtbannister'

import os
import math
from collections import OrderedDict

MPCORB_FILE = os.path.join(os.getenv('HOME', '/Users/bannisterm/'), 'MPCORB-Distant.dat')
L7MODEL = '/Users/bannisterm/Dropbox/OSSOS/Release_summaries/L7model-3.0-9.0'  # 'vos:OSSOS/CFEPS/L7SyntheticModel-v09.txt'
L7_HOME = '/Users/bannisterm/Dropbox/OSSOS/Release_summaries/'
REAL_KBO_AST_DIR = '/Users/bannisterm/Dropbox/OSSOS/measure3/ossin/'
# REAL_KBO_AST_DIR = 'vos:OSSOS/dbaseclone/ast/'
RELEASE_VERSION = 'current'

first_quarter = '/Users/bannisterm/Dropbox/Papers in progress/OSSOS/First_quarter/'
RELEASE_DETECTIONS = {
    4: first_quarter + 'data/v5-oe+u.detections',
    5: L7_HOME + 'v5/OSSOSv5.detections',
    6: L7_HOME + 'v6.prototype.detections'
}

IDX = REAL_KBO_AST_DIR + 'file.idx'  # for local  # 'vos:OSSOS/dbaseclone/idx/file.idx'  # for vos

PLOT_FIELD_EPOCH = 'Jun14.00'  # Jun14.00 ==> '0' days since the New Moon on Jun14

OSSOS_RUNIDS = []
for year in range(13, 17, 1):
    for semester in ['AP', 'BP']:
        for code in ['05', '06']:
            OSSOS_RUNIDS.append('{}{}{}'.format(year, semester, code))

SURVEY_START = '2013-01-01'

PLOT_USNO_STARS = True
PLOT_MEGACAM_ARCHIVE_FIELDS = False  # # TODO Make this work when True
PLOT_SYNTHETIC_KBOS = False
PLOT_SYNTHETIC_KBO_TRAILS = False
PLOT_REAL_KBOS = True and os.access(REAL_KBO_AST_DIR, os.F_OK)
PLOT_FIELD_LAYOUT = True
PLOT_MPCORB = True and os.access(MPCORB_FILE, os.F_OK)

NEWMOONS = {'Feb13': "2013/02/10 10:00:00",
            'Mar13': "2013/03/11 10:00:00",
            'Apr13': "2013/04/10 10:00:00",
            'May13': "2013/05/09 10:00:00",
            'Jun13': "2013/06/08 10:00:00",
            'Jul13': "2013/07/08 10:00:00",
            'Aug13': "2013/08/06 10:00:00",
            'Sep13': '2013/09/05 10:00:00',
            'Oct13': '2013/10/04 10:00:00',
            'Nov13': '2013/11/03 10:00:00',
            'Dec13': '2013/12/02 10:00:00',
            'Jan14': '2014/01/01 10:00:00',
            'Feb14': '2014/01/31 10:00:00',
            'Mar14': '2014/03/28 10:00:00',
            'Apr14': '2014/04/01 10:00:00',
            'May14': '2014/05/28 10:00:00',
            'Jun14': '2014/06/26 10:00:00',
            'Jul14': '2014/07/26 10:00:00',
            'Aug14': "2014/08/25 10:00:00",
            'Sep14': '2014/09/24 10:00:00',
            'Oct14': '2014/10/23 10:00:00',
            'Nov14': '2014/11/22 10:00:00',
            'Dec14': '2014/12/22 10:00:00',
            'Jan15': '2015/01/20 10:00:00',
            'Feb15': '2015/02/18 10:00:00',
            'Mar15': '2015/03/19 10:00:00',
            'Apr15': '2015/04/18 10:00:00',
            'May15': '2015/05/17 10:00:00',
            'Jun15': '2015/06/16 10:00:00',
            'Jul15': '2015/07/15 10:00:00',
            'Aug15': '2015/08/14 10:00:00',
            'Sep15': '2015/09/12 10:00:00',
            'Oct15': '2015/10/12 10:00:00',
            'Nov15': '2015/11/11 10:00:00',
            'Dec15': '2015/12/11 10:00:00',
            'Jan16': '2016/01/09 10:00:00',
            'Feb16': '2016/03/08 10:00:00',
            'Mar16': '2016/03/09 10:00:00',
            'Apr16': '2016/04/08 10:00:00',
            'May16': '2016/05/07 10:00:00',
            'Jun16': '2016/06/05 10:00:00',
            'Jul16': '2016/07/05 10:00:00',
            'Aug16': '2016/08/03 10:00:00',
            'Sep16': '2016/09/01 10:00:00',
            'Oct16': '2016/10/01 10:00:00',
            'Nov16': '2016/11/01 10:00:00',
            'Dec16': '2016/12/01 10:00:00',
            'Jan17': '2017/01/01 10:00:00',
            'Feb17': '2017/02/01 10:00:00',
}

# ORIGINAL: DON't USE, does not match the location of the data as pointed & taken!
# Ablocks={'13AE': {"RA": "14:32:30.29","DEC":"-13:54:01.4"},
# '13AO': {"RA": "16:17:04.41","DEC":"-13:14:45.8"}}
BLOCKS = OrderedDict([
    # E+0+0: image 1616681, ccd21 on April 9. E block discovery triplets are April 4,9,few 19
    ('13AE', {"RA": "14:15:28.89", "DEC": "-12:32:28.4"}),
    ('13AO', {"RA": "15:58:01.35", "DEC": "-12:19:54.2"}),  # O+0+0: image 1625346, ccd21 on May 8. O block are May 7,8.
    ('13BL', {'RA': "00:54:00.00", "DEC": "+03:50:00.00"}),  # 13B blocks are at their opposition locations

    # ('14BH', {'RA': "01:30:00.00", "DEC": "+13:00:00.00"}), # due to bad weather, discovery wasn't until 2014, so 14BH
    ('14BH', {'RA': "01:35:14.39", "DEC": "+13:28:25.2"}),  # where it was actually observed after precession from 2013.

    ('15AP', {'RA': "13:30:00.00", "DEC": "-7:45:00.00"}),  # on-plane
    ('15AM', {'RA': "15:35:00.00", "DEC": "-12:10:00.0"}),  # positioned for its 2015 discovery opposition.
    ('15BS', {'RA': "00:30:00.00", "DEC": "+05:00:00.00"}),
    ('15BD', {'RA': "03:15:00.00", "DEC": "+16:30:00.00"})
])

DISCOVERY_DATES = {"13AE": "2013/04/09 08:50:00",
                   "13AO": "2013/05/08 08:50:00",
                   "15AP": NEWMOONS['Apr15'],
                   "15AM": NEWMOONS['May15'],      # Backup triplet was on 2014/05/29, 2014/06/01 at diff block centre
                   "13BL": "2013/09/29 08:50:00",  # HOWEVER: discovery date is split between months (earliest)
                   "14BH": "2014/10/22 09:30:00",  # Note: Col3N triplet is instead 2014/01/03.
                   "15BS": "2015/09/09 08:50:00",  # both halves within a night: 09-08 and 09-09
                   "15BD": NEWMOONS['Nov15']      # FIXME: set when observations taken
}

DISCOVERY_NEW_MOON = 'Apr13'  # applies to the 13A blocks

OPPOSITION_DATES = {"13AE": NEWMOONS['Apr13'],
                    "13AO": NEWMOONS['May13'],
                    "15AP": NEWMOONS['Apr15'],
                    "15AM": NEWMOONS['May15'],
                    "13BL": NEWMOONS['Oct13'],
                    "14BH": NEWMOONS['Oct13'],
                    "15BS": NEWMOONS['Sep15'],
                    "15BD": NEWMOONS['Nov15'],

}

OSSOS_FILTERS = ['R.MP9601',
                 'R.MP9602',
                 'U.MP9301',
                 'U.MP9302',
                 'GRI.MP9605',
                 ]

COLOSSOS = [
    # the L block in the old money - will need to translate after v5 numbering
    'O13BL3RB',  # o3l39
    'O13BL3S3',  # o3l46
    'O13BL3QX',  # o3l43
    'O13BL3TF',  # o3l76
    'O13BL3TA',  # o3l79
    'O13BL3SN',  # o3l32
    'O13BL3RE',  # o3l77
    'O13BL3SY',  # o3l15
    'O13BL3RN',  # o3l09
    'O13BL3SW',  # o3l13PD
    # 'O13BL3RT', # nfrL3RT : not found in the rerun. Excluded from characterised sample!
    'O13BL3RQ',  # o3l01
    'O13BL3SH',  # o3l06
    'O13BL3R1',  # o3l18
    'O13BL3RH',  # o3l57

    # the H block so far has some
    'Col3N01',
    'Col3N02',
    'Col3N03',
    'Col3N05',
    'Col3N06',
    'Col3N07',

    # the O and E blocks ColOSSOS, will only ever be in these designations
    'o3e11',
    'o3e38',
    'o3e45',  # stirred object
    'o3e25',
    'o3e34PD',
    'o3o18',
    'o3e31',
    'o3e49',
    'o3o14',
    'o3e32',
    'o3o21',
    'o3e53',
    'o3o15',
    'o3e35',
    'o3e16',
    'o3e29',
    'o3e43',
    'o3o28',
    'o3e55',
    'o3o27',
    'o3e19',
    'o3e21',
    'o3e04',
    'o3o01',
    'o3e37PD',
    'o3e23PD',
    'o3e02',
    'o3e44',
    'o3o29',
    'o3o09',
    'o3e27PD',
    'o3e20PD',
    'o3e39',
    'o3e30PD',
    'o3e22',
    'o3o11',
    'o3e09',
    'o3o34',
    'o3e05',
    'o3e01',
    'o3o20PD',
]


class VersionError(Exception):
    'An error with the TNO summary-file version'


class tno(object):
    ''' This takes two types of input lines, which populate almost the same but subtly different amounts of information.
        Standard orbital elements are always populated.
        Combining both class methods is the most reliable way of instantiating these objects from a data release.
    '''

    def __init__(self, name, dist, dist_e, nobs, arclen, av_xres, av_yres, max_x, max_y,
                 a, a_e, e, e_e, i, i_e, node, node_e, argperi, argperi_e, time_peri, time_peri_e):
        # self.classification = str(classification)
        self.name = str(name)
        self.dist = float(dist)
        self.dist_e = float(dist_e)
        self.nobs = int(nobs)
        self.arclen = float(arclen)
        self.av_xres = float(av_xres)
        self.av_yres = float(av_yres)
        self.max_x = float(max_x)
        self.max_y = float(max_y)
        self.a = float(a)
        self.a_e = float(a_e)
        self.e = float(e)
        self.e_e = float(e_e)
        self.i = float(i)
        self.i_e = float(i_e)
        self.node = float(node)
        self.node_e = float(node_e)
        self.argperi = float(argperi)
        self.argperi_e = float(argperi_e)
        self.time_peri = float(time_peri)
        self.time_peri_e = float(time_peri_e)
        self.peri = self.a * (1. - self.e)

    @classmethod
    def from_summary_line(cls, summaryLine, version=4, existing_object=None):
        ''' Summary format:
        object  mag  stdev   dist  ..E nobs time av_xres av_yres max_x max_y
        a ..E  e  ..E  i ..E    node     ..E    argperi     ..E           M        ..E   ra_dis  dec_dis
        '''
        if not summaryLine:
            raise ValueError('No summary line given')
        if version == 4:
            params = summaryLine.split()
            if len(params) != 25:
                print params
                raise TypeError('Expected 25 columns, {0} given'.format(len(params)))
            input_params = params[0:1] + params[3:23]
            if not existing_object:
                retval = cls(*input_params)
            else:
                assert isinstance(existing_object, tno)
                assert existing_object.name == params[0]
                retval = existing_object
            retval.mean_mag = float(params[1])
            retval.mean_mag_stdev = float(params[2])
            retval.ra_discov = float(params[23])
            retval.dec_discov = float(params[24])
        else:
            raise VersionError('Unknown version "{0}"'.format(version))
        assert retval
        return retval

    @classmethod
    def from_class_line(cls, classLine, version=4, existing_object=None):
        '''
        Class format:
        class wrt n m security object  mag  stdev F H_sur  dist     ..E    nobs   time  av_xres av_yres max_x  max_y
        a ..E        e  ..E      i      ..E      node     ..E   argperi      ..E   time_peri      ..E    rate
        '''
        if not classLine:
            raise ValueError('No class file line given.')
        if version == 4:
            params = classLine.split()
            if len(params) != 31:
                print params
                raise TypeError('Expected 31 columns, {0} given'.format(len(params)))
            input_params = [params[5]] + params[10:30]  # the elements that are in common
            if not existing_object:
                # can later add more tests that the values match those in the existing object
                retval = cls(*input_params)
            else:
                assert isinstance(existing_object, tno)
                assert existing_object.name == params[5]
                retval = existing_object
            retval.classification = params[0]
            retval.wrt = params[1]
            retval.n = int(params[2])
            retval.m = int(params[3])
            retval.security = params[4]
            retval.mag_discov = float(params[6])
            retval.mag_discov_e = float(params[7])
            retval.filter = params[8]
            retval.H = float(
                params[9])  # H_r magnitude, using the average m_r and the discovery geometry (same error as m_r)
            retval.rate = float(params[30])
        elif version == 5:
            # Upgrade?
            # input_params =
            retval = cls(*input_params)
        else:
            raise VersionError('Unknown version "{0}"'.format(version))
        assert retval
        return retval


def mag_at_radius(r, d, p=0.10, phi=1):
    # Assumptions: m_r,
    #                  small-object albedo,
    #                  observation at opposition,
    #                  heliocentric and geocentric distance approximate as equal at TNO distances.
    # r in km
    # d in AU
    m_sun = -27.1

    m_r = m_sun - 2.5 * math.log10((p * phi * r ** 2) / (2.25 * 10 ** 16 * (d ** 2) * (d - 1) ** 2))
    print "m_r = {:2.2f} for a {} km radius TNO at {} AU at opposition, assuming {} albedo.".format(
        m_r, r, d, p)

    return m_r


if __name__ == '__main__':
    # want m_r = 24.5: plutinos are about 10% albedo
    mag_at_radius(20, 30)



__author__ = "David Rusk <drusk@uvic.ca>"

import math

from hamcrest.core.base_matcher import BaseMatcher


def round_sigfigs(x, sigfigs):
    return round(x, -int(math.floor(math.log10(math.fabs(x)))) + (sigfigs - 1))


class IsAlmostEqualSigFigs(BaseMatcher):
    def __init__(self, expected, sigfigs):
        self.expected = expected
        self.sigfigs = sigfigs

    def _matches(self, actual):
        return (round_sigfigs(self.expected, self.sigfigs) ==
                round_sigfigs(actual, self.sigfigs))

    def describe_to(self, description):
        description.append_text(
            "expected: %f to %d "
            "significant figures" % (self.expected, self.sigfigs))


def almost_equal(expected, sigfigs):
    return IsAlmostEqualSigFigs(expected, sigfigs)



#!/usr/bin/env python



import urllib
from astropy.io import votable

def TAPQuery(RAdeg, DECdeg, width, height):
    """Given the RA/DEC query the USNO2B catalog for stars insize box at that location.

    returns a file-like object

    """


    QUERY=( """ SELECT """
            """ "I/322A/out".RAJ2000, """
            """ "I/322A/out".DEJ2000, """
            """ "I/322A/out".Bmag, """
            """ "I/322A/out".Vmag, """
            """ "I/322A/out".rmag """
            """ FROM "I/322A/out" """
            """ WHERE 1=CONTAINS(POINT('ICRS',"I/322A/out".RAJ2000, "I/322A/out".DEJ2000), BOX('ICRS', %f, %f, %f, %f)) AND "I/322A/out".Bmag < 12 """ % ( RAdeg, DECdeg, width, height))
    
    
    data={"query": QUERY,
          "request": "doQuery",
          "lang": "ADQL",
          "format": "votable"}
    
    
    url="http://tapvizier.u-strasbg.fr/TAPVizieR/tap/sync"
    
    return urllib.urlopen(url, urllib.urlencode(data))




if __name__ == '__main__':
    ra_cen = (37+7)/2.0
    dec_cen = (20+0)/2.0
    width = 20-0
    height = 37-7
    print votable.parse(TAPQuery(ra_cen, dec_cen, width, height).read()).get_first_table()
    



# !python
# Copyright 2012, 2013 JJ Kavelaars

import argparse
import tempfile
import math
import sys
import time

import requests
import ephem
from astropy.io.votable import parse
import matplotlib

from ossos import storage


matplotlib.use('Agg')
from matplotlib.pyplot import figure, close
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_pdf import PdfPages
import logging

from ossos.planning.plotting import parameters


def query_for_observations(mjd, observable, runids):
    """Do a QUERY on the TAP service for all observations that are part of runid, 
    where taken after mjd and have calibration 'observable'.

    Schema is at: http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/tap/tables

    mjd : float 
    observable: str ( 2 or 1 )
    runid: tuple eg. ('13AP05', '13AP06')

    """

    data = {"QUERY": ("SELECT Observation.target_name as TargetName, "
                      "COORD1(CENTROID(Plane.position_bounds)) AS RA,"
                      "COORD2(CENTROID(Plane.position_bounds)) AS DEC, "
                      "Plane.time_bounds_cval1 AS StartDate, "
                      "Plane.time_exposure AS ExposureTime, "
                      "Observation.instrument_name AS Instrument, "
                      "Plane.energy_bandpassName AS Filter, "
                      "Observation.observationID AS dataset_name, "
                      "Observation.proposal_id AS ProposalID, "
                      "Observation.proposal_pi AS PI "
                      "FROM caom2.Observation AS Observation "
                      "JOIN caom2.Plane AS Plane ON "
                      "Observation.obsID = Plane.obsID "
                      "WHERE  ( Observation.collection = 'CFHT' ) "
                      "AND Plane.time_bounds_cval1 > %d "
                      "AND Plane.calibrationLevel=%s "
                      "AND Observation.proposal_id IN %s " ) %
                     ( mjd, observable, str(runids)),
            "REQUEST": "doQuery",
            "LANG": "ADQL",
            "FORMAT": "votable"}

    result = requests.get(storage.TAP_WEB_SERVICE, params=data, verify=False)
    assert isinstance(result, requests.Response)
    logging.debug("Doing TAP Query using url: %s" % (str(result.url)))
    tmpFile = tempfile.NamedTemporaryFile()
    with open(tmpFile.name, 'w') as outfile:
        outfile.write(result.text)
    vot = parse(tmpFile.name).get_first_table()

    vot.array.sort(order='StartDate')
    t = vot.array
    tmpFile.close()

    print t

    logging.debug("Got {} lines from tap query".format(len(t)))

    return t


def create_ascii_table(obsTable, outfile):
    """Given a table of observations create an ascii log file for easy parsing.
    Store the result in outfile (could/should be a vospace dataNode)

    obsTable: astropy.votable.array object
    outfile: str (name of the vospace dataNode to store the result to)

    """

    logging.info("writing text log to %s" % ( outfile))

    stamp = "#\n# Last Updated: " + time.asctime() + "\n#\n"
    header = "| %20s | %20s | %20s | %20s | %20s | %20s | %20s |\n" % (
        "EXPNUM", "OBS-DATE", "FIELD", "EXPTIME(s)", "RA", "DEC", "RUNID")
    bar = "=" * (len(header) - 1) + "\n"

    if outfile[0:4] == "vos:":
        tmpFile = tempfile.NamedTemporaryFile(suffix='.txt')
        fout = tmpFile
    else:
        fout = open(outfile, 'w')

    t2 = None
    fout.write(bar + stamp + bar + header)

    populated = storage.list_dbimages()
    for i in range(len(obsTable) - 1, -1, -1):
        row = obsTable.data[i]
        if row['dataset_name'] not in populated:
            storage.populate(row['dataset_name'])
        sDate = str(ephem.date(row.StartDate +
                               2400000.5 -
                               ephem.julian_date(ephem.date(0))))[:20]
        t1 = time.strptime(sDate, "%Y/%m/%d %H:%M:%S")
        if t2 is None or math.fabs(time.mktime(t2) - time.mktime(t1)) > 3 * 3600.0:
            fout.write(bar)
        t2 = t1
        ra = str(ephem.hours(math.radians(row.RA)))
        dec = str(ephem.degrees(math.radians(row.DEC)))
        line = "| %20s | %20s | %20s | %20.1f | %20s | %20s | %20s |\n" % (
            str(row.dataset_name),
            str(ephem.date(row.StartDate + 2400000.5 -
                           ephem.julian_date(ephem.date(0))))[:20],
            row.TargetName[:20],
            row.ExposureTime, ra[:20], dec[:20], row.ProposalID[:20] )
        fout.write(line)

    fout.write(bar)

    if outfile[0:4] == "vos:":
        fout.flush()
        storage.copy(tmpFile.name, outfile)
    fout.close()

    return


def create_sky_plot(obstable, outfile, night_count=1, stack=True):
    """Given a VOTable that describes the observation coverage provide a PDF of the skycoverge.

    obstable: vostable.arrary
    stack: BOOL (true: stack all the observations in a series of plots)

    """

    # camera dimensions
    width = 0.98
    height = 0.98

    if outfile[0:4] == 'vos:':
        tmpFile = tempfile.NamedTemporaryFile(suffix='.pdf')
        pdf = PdfPages(tmpFile.name)
    else:
        pdf = PdfPages(outfile)

    saturn = ephem.Saturn()
    uranus = ephem.Uranus()

    t2 = None
    fig = None
    proposalID = None
    limits = {'13A': ( 245, 200, -20, 0),
              '13B': ( 0, 45, 0, 20)}
    for row in reversed(obstable.data):
        date = ephem.date(row.StartDate + 2400000.5 - ephem.julian_date(ephem.date(0)))
        sDate = str(date)
        # Saturn only a problem in 2013A fields
        saturn.compute(date)
        sra = math.degrees(saturn.ra)
        sdec = math.degrees(saturn.dec)
        uranus.compute(date)
        ura = math.degrees(uranus.ra)
        udec = math.degrees(uranus.dec)
        t1 = time.strptime(sDate, "%Y/%m/%d %H:%M:%S")
        if t2 is None or (math.fabs(time.mktime(t2) - time.mktime(
                t1)) > 3 * 3600.0 and opt.stack) or proposalID is None or proposalID != row.ProposalID:
            if fig is not None:
                pdf.savefig()
                close()
            proposalID = row.ProposalID
            fig = figure(figsize=(7, 2))
            ax = fig.add_subplot(111, aspect='equal')
            ax.set_title("Data taken on %s-%s-%s" % ( t1.tm_year, t1.tm_mon, t1.tm_mday), fontdict={'fontsize': 8})
            ax.axis(limits.get(row.ProposalID[0:3], (0, 20, 0, 20)))  # appropriate only for 2013A fields
            ax.grid()
            ax.set_xlabel("RA (deg)", fontdict={'fontsize': 8})
            ax.set_ylabel("DEC (deg)", fontdict={'fontsize': 8})
        t2 = t1
        ra = row.RA - width / 2.0
        dec = row.DEC - height / 2.0
        color = 'b'
        if 'W' in row['TargetName']:
            color = 'g'
        ax.add_artist(Rectangle(xy=(ra, dec), height=height, width=width,
                                edgecolor=color, facecolor=color,
                                lw=0.5, fill='g', alpha=0.33))
        ax.add_artist(Rectangle(xy=(sra, sdec), height=0.3, width=0.3,
                                edgecolor='r',
                                facecolor='r',
                                lw=0.5, fill='k', alpha=0.33))
        ax.add_artist(Rectangle(xy=(ura, udec), height=0.3, width=0.3,
                                edgecolor='b',
                                facecolor='b',
                                lw=0.5, fill='b', alpha=0.33))

    if ax is not None:
        ax.axis((270, 215, -20, 0))
        pdf.savefig()
        close()
    pdf.close()
    if outfile[0:4] == "vos:":
        storage.copy(tmpFile.name, outfile)
        tmpFile.close()

    return


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Query the CADC for OSSOS observations.")

    parser.add_argument('date', nargs='?', action='store',
                        default=parameters.SURVEY_START)

    parser.add_argument('--runid', nargs='*', action='store',
                        default=parameters.OSSOS_RUNIDS)

    parser.add_argument('--cal', action='store', default=1)

    parser.add_argument('--outfile', action='store',
                        default='vos:OSSOS/ObservingStatus/obsList')

    parser.add_argument('--debug', action='store_true')

    parser.add_argument('--stack', action='store_true', default=False,
                        help=( "Make single status plot that stacks"
                               " data accross multiple nights, instead of nightly sub-plots." ))

    opt = parser.parse_args()

    runids = tuple(opt.runid)

    if opt.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)

    try:
        mjd_yesterday = ephem.date(ephem.julian_date(ephem.date(opt.date))) - 2400000.5
    except Exception as e:
        logging.error("you said date = %s" % (opt.date))
        logging.error(str(e))
        sys.exit(-1)

    obs_table = query_for_observations(mjd_yesterday, opt.cal, runids)

    create_ascii_table(obs_table, opt.outfile + ".txt")

    # create_sky_plot(obs_table, opt.outfile + ".pdf", stack=opt.stack)



"""OSSOS helper methods"""
from datetime import datetime
from logging import handlers
import logging
import numpy
import os
import re
import subprocess
import time
from vos import vos
import sys

try:
    from astropy.time import erfa_time
except ImportError:
    from astropy.time import sofa_time as erfa_time
from astropy.time import TimeString
from astropy.time import Time


MATCH_TOLERANCE = 100.0


def task():
    return os.path.splitext(os.path.basename(sys.argv[0]))[0]


def exec_prog(args):
    """Run a subprocess, check for .OK and raise error if does not exist.

    args:  list of arguments, for value is the command to execute.
    """

    program_name = args[0]
    logging.info(" ".join(args))
    output = subprocess.check_output(args, stderr=subprocess.STDOUT)
    if not os.access(program_name+".OK", os.F_OK):
        logging.error("No {}.OK file?".format(program_name))
        raise subprocess.CalledProcessError(-1, ' '.join(args), output)
    os.unlink(program_name+".OK")
    if os.access(program_name+".FAILED", os.F_OK):
        os.unlink(program_name+".FAILED")
    return output


class VOFileHandler(handlers.BufferingHandler):
    """
    A handler class that writes formatted logging records to VOSpace files.
    """
    def __init__(self, filename, vos_client=None):
        self.filename = filename
        self._client = vos_client
        self.stream = None
        super(VOFileHandler, self).__init__(1024*1024)

    @property
    def client(self):
        """
        Send back the client we were sent, or construct a default one.
        """
        if self._client is not None:
            return self._client
        self._client = vos.Client()
        return self._client

    def close(self):
        """
        Closes the stream.
        """
        self.flush()
        try:
            if self.stream is not None:
                self.stream.close()
                self.stream = None
        except:
            pass

    def flush(self):
        if self.stream is None:
            self.stream = self._open()
        for record in self.buffer:
            self.stream.write("{}\n".format(self.format(record)))
        self.buffer = []

    def _open(self):
        """
        Open the current base file with the (original) mode
        Return the resulting stream.
        """
        try:
            if self.stream.closed:
                return self.stream
        except AttributeError:
            pass

        buff = None
        try:
            buff = self.client.open(self.filename, view='data').read()
        except:
            pass

        stream = self.client.open(self.filename, os.O_WRONLY)
        if buff is not None:
            stream.write(buff)

        return stream


def get_pixel_bounds_from_datasec_keyword(datasec):
    """
    Return the x/y pixel boundaries of the data section.
    :param datasec: str e.g. '[33:2080,1:4612]'
    :return: ((xmin,xmax),(ymin,ymax))
    """
    datasec = re.findall(r'(\d+)', datasec)
    x1 = min(int(datasec[0]), int(datasec[1]))
    x2 = max(int(datasec[0]), int(datasec[1]))
    y1 = min(int(datasec[2]), int(datasec[3]))
    y2 = max(int(datasec[2]), int(datasec[3]))

    return (x1, x2), (y1, y2)


def match_lists(pos1, pos2, tolerance=MATCH_TOLERANCE):
    """
    Given two sets of x/y positions match the lists, uniquely.

    :rtype : numpy.ma, numpy.ma
    :param pos1: list of x/y positions.
    :param pos2: list of x/y positions.
    :param tolerance: float distance, in pixels, to consider a match

    Algorithm:
        - Find all the members of pos2 that are within tolerance of pos1[idx1].
                These pos2 members are match_group_1
        - Find all the members of pos1 that are within tolerance of match_group_1[idx2].
                These pos1 members are match_group_2
        - If pos1[idx] is in match_group_2 then pos1[idx] is a match of object at match_group_1[idx2]

    """

    assert isinstance(pos1, numpy.ndarray)
    assert isinstance(pos2, numpy.ndarray)

    # build some arrays to hold the index of things that matched between lists.
    npts1 = len(pos1[:, 0])
    pos1_idx_array = numpy.arange(npts1, dtype=numpy.int16)
    npts2 = len(pos2[:, 0])
    pos2_idx_array = numpy.arange(npts2, dtype=numpy.int16)

    # this is the array of final matched index, -1 indicates no match found.
    match1 = numpy.ma.zeros(npts1, dtype=numpy.int16)
    match1.mask = True

    # this is the array of matches in pos2, -1 indicates no match found.
    match2 = numpy.ma.zeros(npts2, dtype=numpy.int16)
    match2.mask = True

    for idx1 in range(npts1):

        # compute the distance source idx1 to each member of pos2
        sep = numpy.sqrt((pos2[:, 0] - pos1[idx1, 0]) ** 2 + (pos2[:, 1] - pos1[idx1, 1]) ** 2)

        # considered a match if sep is below tolerance and is the closest match available.
        match_condition = numpy.all((sep <= tolerance, sep == sep.min()), axis=0)

        # match_group_1 is list of the indexes of pos2 entries that qualified as possible matches to pos1[idx1]
        match_group_1 = pos2_idx_array[match_condition]

        # For each of those pos2 objects that could be a match to pos1[idx] find the best match in all of pos1
        for idx2 in match_group_1:
            # compute the distance from this pos2 object that is a possible match to pos1[idx1] to all members of pos1
            sep = numpy.sqrt((pos1[:, 0] - pos2[idx2, 0]) ** 2 + (pos1[:, 1] - pos2[idx2, 1]) ** 2)

            # considered a match if sep is below tolerance and is the closest match available.
            match_condition = numpy.all((sep <= tolerance, sep == sep.min()), axis=0)
            match_group_2 = pos1_idx_array[match_condition]

            # Are any of the pos1 members that were matches to the matched pos2 member the pos1[idx] entry?
            if idx1 in match_group_2:
                match1[idx1] = idx2
                match2[idx2] = idx1
                # this BREAK is in here since once we have a match we're done.
                break

    return match1, match2


class TimeMPC(TimeString):
    """
    Override the TimeString class to convert from MPC format string to astropy.time.Time object.

    usage:

    from astropy.time.core import Time
    Time.FORMATS[TimeMPC.name] = TimeMPC

    t = Time('2000 01 01.00001', format='mpc', scale='utc')

    str(t) == '2000 01 01.000001'
    """

    name = 'mpc'
    subfmts = (('mpc', '%Y %m %d', "{year:4d} {mon:02d} {day:02d}.{fracday:s}"),)

    def __init__(self, val1, val2, scale, precision=6,
                 in_subfmt=None, out_subfmt=None, from_jd=False):
        super(TimeMPC, self).__init__(val1=val1,
                                      val2=val2,
                                      scale=scale,
                                      precision=precision,
                                      in_subfmt=in_subfmt,
                                      out_subfmt=out_subfmt,
                                      from_jd=from_jd)
        self.precision = precision

    # ## need our own 'set_jds' function as the MPC Time string is not typical
    def set_jds(self, val1, val2):
        """

        Parse the time strings contained in val1 and set jd1, jd2

        :param val1: array of strings to parse into JD format
        :param val2: not used for string conversions but passed regardless
        """

        # This routine is based on atropy.time.core.TimeString class.

        iterator = numpy.nditer([val1, None, None, None, None, None, None],
                                op_dtypes=[val1.dtype] + 5 * [numpy.intc] + [numpy.double])
        subfmts = self.subfmts
        for val, iy, im, iday, ihr, imin, dsec in iterator:
            time_str = val.item()

            # Assume that anything following "." on the right side is a
            # floating fraction of a day.
            try:
                idot = time_str.rindex('.')
            except:
                fracday = 0.0
            else:
                time_str, fracday = time_str[:idot], time_str[idot:]
                fracday = float(fracday)

            for _, strptime_fmt, _ in subfmts:
                try:
                    tm = time.strptime(time_str, strptime_fmt)
                except ValueError:
                    pass
                else:
                    iy[...] = tm.tm_year
                    im[...] = tm.tm_mon
                    iday[...] = tm.tm_mday
                    hrprt = tm.tm_hour + int(24 * fracday)
                    ihr[...] = hrprt
                    mnprt = tm.tm_min + int(60 * (24 * fracday - hrprt))
                    imin[...] = mnprt
                    dsec[...] = tm.tm_sec + 60 * (60 * (24 * fracday - hrprt) - mnprt)

                    break
            else:
                raise ValueError("Time {0} does not match {1} format".format(time_str, self.name))

        self.jd1, self.jd2 = erfa_time.dtf_jd(
            self.scale.upper().encode('utf8'), *iterator.operands[1:])

        return

    def str_kwargs(self):
        """
        Generator that yields a dict of values corresponding to the
        calendar date and time for the internal JD values.
        """
        iys, ims, ids, ihmsfs = erfa_time.jd_dtf(self.scale.upper()
                                                 .encode('utf8'),
                                                 6,
                                                 self.jd1, self.jd2)

        # Get the str_fmt element of the first allowed output subformat
        _, _, str_fmt = self._select_subfmts(self.out_subfmt)[0]

        yday = None
        has_yday = '{yday:' in str_fmt or False

        ihrs = ihmsfs[..., 0]
        imins = ihmsfs[..., 1]
        isecs = ihmsfs[..., 2]
        ifracs = ihmsfs[..., 3]
        for iy, im, iday, ihr, imin, isec, ifracsec in numpy.nditer(
                [iys, ims, ids, ihrs, imins, isecs, ifracs]):
            if has_yday:
                yday = datetime(iy, im, iday).timetuple().tm_yday

            fracday = (((((ifracsec / 1000000.0 + isec) / 60.0 + imin) / 60.0) + ihr) / 24.0) * (10 ** 6)
            fracday = '{0:06g}'.format(fracday)[0:self.precision]

            yield {'year': int(iy), 'mon': int(im), 'day': int(iday),
                   'hour': int(ihr), 'min': int(imin), 'sec': int(isec),
                   'fracsec': int(ifracsec), 'yday': yday, 'fracday': fracday}

Time.FORMATS['mpc'] = TimeMPC



#!/usr/bin/python
import sys
import math
import string

D2PI = 6.2831853071795864769252867665590057683943387987502
DS2R = 7.2722052166430399038487115353692196393452995355905e-5
MJD0 = 51544.5
DPY = 365.25


def mjd2gmst(mjd):
    """Convert Modfied Juian Date (JD = 2400000.5) to GMST
    
    Taken from P.T. Walace routines.

    """

    tu = (mjd - MJD0) / (100*DPY)

    st = math.fmod(mjd, 1.0) * D2PI + (24110.54841 + (8640184.812866 + (0.093104 - 6.2e-6 * tu) * tu) * tu) * DS2R

    w = math.fmod(st, D2PI)
    if w >= 0.0:
        return w
    else:
        return w + D2PI


class Coord:
    """A class to manipulate astronomical coordinates.

    """

    def __init__(self, (l, b), system='ICRS'):
        """Assign the RA/DEC of the coordinate (initialization).

        """

        self.system = system
        if system == 'ecliptic':
            self.el = l
            self.eb = b
            self.ec2eq()
        elif system == 'ICRS':
            self.ra = l
            self.dec = b
            self.eq2ec()
        else:
            sys.stderr.write("Don't know this system: {}".format(system))
            sys.stderr.write("Use ICRS or ecliptic")
            sys.exit(-1)

    def set_system(self, system):
        self.system = system

    def zd(self, phi):
        """Compute the Zennith Distance of this coord at given OBS.

        """

        return math.sin(phi)

    def __segs__(self, ang, precision=1):
        """given an angle, convert it to a segisdecimal string"""
        sign = "+"
        if ang < 0:
            sign = "-"
            ang = ang * -1.0000000000000
        ang += 1E-11

        _d = math.floor(ang)
        _mf = (ang - _d) * 60.00
        _m = math.floor(_mf)
        _s = (_mf - _m) * 60.00
        _is = math.floor(_s)
        _fs = math.floor((_s - _is) * 10.00 * precision)

        _d = "%.0f" % ( _d)
        _m = "%.0f" % ( _m)
        _is = "%.0f" % (_is)
        _fs = int(_fs)

        _sd = string.zfill(_d, 2)
        _sm = string.zfill(_m, 2)
        _ss = string.zfill(_is, 2)
        _sf = string.zfill(_fs, precision)

        s = sign + "%s:%s:%s.%s" % ( _sd, _sm, _ss, _sf)
        return s

    def __str__(self):
        """String representation of coordinate.

        """

        (l, b) = (0, 0)
        if self.system == 'ICRS':
            (l, b) = (self.ra, self.dec)
        elif self.system == 'ecliptic':
            (l, b) = (self.el, self.eb)
        l = math.degrees(l / 15.E0)
        b = math.degrees(b)
        return self.__segs__(l, 2) + " " + self.__segs__(b, 1)

    def ec2eq(self):
        """Convert ecliptic coordinates to equatorial coordinates"""

        sin = math.sin
        cos = math.cos
        arcsin = math.asin
        arccos = math.acos

        eb = self.eb
        el = self.el
        ob = math.radians(23.439281)

        dec = arcsin(sin(eb) * cos(ob) + cos(eb) * sin(ob) * sin(el))
        sra = (sin(dec) * cos(ob) - sin(eb)) / (cos(dec) * sin(ob))

        cra = cos(el) * cos(eb) / cos(dec)
        if sra < 1 and sra > -1:
            sa = arcsin(sra)
        else:
            sa = 0
        ca = arccos(cra)
        tsa = sa
        tca = ca
        if tsa < 0:
            ca = 2.0 * math.pi - ca
        if tca >= math.pi / 2.0:
            sa = math.pi - sa
        if ca >= math.pi * 2.0:
            ca = ca - math.pi * 2.0
        self.tsa = sra
        self.tca = cra
        self.ra = ca
        self.dec = dec

    def eq2ec(self):

        sin = math.sin
        cos = math.cos
        arcsin = math.asin
        arccos = math.acos

        ra = self.ra
        dec = self.dec
        ob = math.radians(23.439281000000000000)

        eb = arcsin(sin(dec) * cos(ob) - cos(dec) * sin(ra) * sin(ob))
        sl = (sin(dec) - sin(eb) * cos(ob)) / (cos(eb) * sin(ob))
        cl = cos(ra) * cos(dec) / cos(eb)
        if sl < -1: sl = float(-1)
        if sl > 1: sl = float(1)
        els = arcsin(sl)
        elc = arccos(cl)
        ela = 0
        elb = 0
        if sl < 0 and cl < 0:
            ela = math.pi * 1 - els
            elb = 2.0 * math.pi - elc
            rule = 0
        if sl < 0 and cl > 0:
            ela = 2.0 * math.pi + els
            elb = 2.0 * math.pi - elc
            rule = 1
        if sl > 0 and cl > 0:
            ela = els
            elb = elc
            rule = 2
        if sl > 0 and cl < 0:
            ela = math.pi - els
            elb = elc
            rule = 3

        self.sl = els
        self.cl = elc
        self.eb = eb
        self.el = ela


if __name__ == '__main__':

    c = 0
    for a in range(0, 360, 9):
        j = Coord((math.radians(a), math.radians(33.0)))
        k = Coord((j.el, j.eb), system='ecliptic')
        k.set_system('ICRS')
        if str(j) != str(k):
            c = c + 1

    if c == 0:
        print "Passed"



from unittest import TestCase

__author__ = 'jjk'

import storage


class TestReset_datasec(TestCase):
    """
    Test that the logic of the storage.reset_datasec method is working correctly.
    """

    def test_reset_datasec_not_flipped(self):
        datasec = "[100:2000,100:4450]"
        cutout = "[101:200, 101:200]"
        naxis1 = 2048
        naxis2 = 4600
        self.assertEquals(storage.reset_datasec(cutout, datasec, naxis1, naxis2), "[1:1900,1:4350]")

    def test_reset_datasec_flipped(self):
        datasec = "[100:2000,100:4450]"
        cutout = "[200:101, 200:101]"
        naxis1 = 2048
        naxis2 = 4600
        self.assertEquals(storage.reset_datasec(cutout, datasec, naxis1, naxis2), "[1:101,1:101]")

    def test_reset_datasec_flipped_stars(self):
        datasec = "[100:2000,100:4450]"
        cutout = "[-*, -*]"
        naxis1 = 2048
        naxis2 = 4600
        self.assertEquals(storage.reset_datasec(cutout, datasec, naxis1, naxis2), "[49:1949,151:4501]")

    def test_reset_datasec_stars(self):
        datasec = "[100:2000,100:4450]"
        cutout = "[*, *]"
        naxis1 = 2048
        naxis2 = 4600
        self.assertEquals(storage.reset_datasec(cutout, datasec, naxis1, naxis2), "[100:2000,100:4450]")

    def test_get_flipped_image(self):
        """
        This test requires VOSpace be accessible.
        """
        storage.DBIMAGES = 'vos:OSSOS/TEST/dbimages/'
        image = storage.get_image('1667740', ccd=1, return_file=False)
        self.assertEqual(image[0].header['DATASEC'],"[33:2080,33:4644]")


from astropy import units

__author__ = "David Rusk <drusk@uvic.ca>"
from baseviewer import WxMPLFitsViewer
from displayable import DisplayableImageSinglet
from interaction import Signal
from ..gui import logger
from ..downloads.cutouts.source import SourceCutout


class SingletViewer(WxMPLFitsViewer):
    """
    Displays a single FITS image at a time.
    """

    def __init__(self, parent, display):
        logger.debug("Bilding {}".format(self))
        super(SingletViewer, self).__init__(parent, display)
        self.xy_changed = Signal()

    def _refresh_markers(self, cutout):
        self._displayables_by_cutout[cutout].clear_markers(self.ds9)
        self.mark_apertures(self.current_cutout)

    def mark_apertures(self, cutout, pixel=False):
        """

        :param cutout: A SourceCutout that is displayed
        :type cutout: SourceCutout
        :param pixel: Mark based on pixel locations or based on RA/DEC ?
        :type pixel: bool
        """
        try:
            x = cutout.reading.mpc_observation.comment.x
            y = cutout.reading.mpc_observation.comment.y
            (x, y) = cutout.get_pixel_coordinates((x, y))
            pixel = True
        except Exception as ex:
            x, y = cutout.pixel_x, cutout.pixel_y

        (ra, dec) = cutout.pix2world(x, y, usepv=False)

        if not self.mark_source:
            return

        try:
            # radii = (cutout.apcor.aperture, cutout.apcor.sky, cutout.apcor.swidth+cutout.apcor.sky)
            radii = (cutout.apcor.sky, cutout.apcor.swidth+cutout.apcor.sky)
        except Exception as ex:
            logger.info("Exception: {0}".format(ex))
            logger.warning("Failed trying to get apcor radius values, using defaults for marking.")
            radii = (15, 30)

        if pixel:
            radii = [radius * 0.185 * units.arcsec for radius in radii]
            self._displayables_by_cutout[cutout].place_annulus(ra, dec, radii, colour='r')
        else:
            cutout.update_pixel_location((x, y), extno=cutout.original_observed_ext)
            radii = [radius * 0.185 * units.arcsec for radius in radii]
            self._displayables_by_cutout[cutout].place_annulus(cutout.ra, cutout.dec, radii, colour='r')

    def place_marker(self, cutout, x, y, radius, colour):
        self._displayables_by_cutout[cutout].place_marker(x, y, radius, colour=colour)

    def register_xy_changed_event_handler(self, handler):
        self.xy_changed.connect(handler)

    def _create_displayable(self, cutout):
        return DisplayableImageSinglet(cutout.hdulist, self.ds9)

    def _attach_handlers(self, displayable):
        displayable.xy_changed.connect(self.xy_changed.fire)
        displayable.focus_released.connect(self.release_focus)

    def _detach_handlers(self, displayable):
        if displayable is not None:
            displayable.xy_changed.disconnect(self.xy_changed.fire)
            displayable.focus_released.disconnect(self.release_focus)



#!python
"""Retrieval of cutouts of the FITS images associated with the OSSOS detections.

Takes a directory of .ast file (in dbase format) as input

An Example URL for cutouts
http://www.canfar.phys.uvic.ca/vospace/auth/synctrans?TARGET=vos://cadc.nrc.ca~
vospace/OSSOS/dbimages/1625356/1625356p.fits&DIRECTION=pullFromVoSpace&PROTOCOL=
ivo://ivoa.net/vospace/core%23httpget&view=cutout&cutout=CIRCLE+ICRS+242.1318+-1
2.4747+0.05
"""

import argparse
import logging
import os

from astropy import units
from astropy.units import Quantity
from astropy.io import fits
from ossos import mpc
from ossos import storage
from planning.plotting import parameters


def cutout(obj, obj_dir, radius):
    print len([n for n in obj.mpc_observations if not n.null_observation])
    for obs in obj.mpc_observations:
        print('starting analysis of {}'.format(str(obs)))
        if obs.null_observation:
            print('skipping')
            continue
        if obs.date > parameters.SURVEY_START:  # can't make postage stamps of earlier linkages
            # can't parse for an obs.comment's exposure number if no obs.comment exists
            try:
                expnum = obs.comment.frame.split('p')[0].strip(' ')  # only want calibrated images
            except AttributeError, e:
                print('No comment in this MPC line!')
                continue
            if not expnum.isdigit():
                print('expnum {} parsed from comment line invalid. Check comment parsing.\n{}'.format(expnum, str(
                    obs.comment)))
                continue
            uri = storage.get_uri(expnum)
            sky_coord = obs.coordinate  # Using the WCS rather than the X/Y (X/Y can be unreliable over the whole survey)
            print('Trying {} on {} on {}...'.format(obj.provisional_name, obs.date, expnum))
            try:
                hdulist = storage.ra_dec_cutout(uri, sky_coord, radius)
                # if not 'ASTLEVEL' in hdulist[1].header:   # FIXME: activate once all headers are retro-fitted with ASTLEVEL
                #     logging.info('Cutout invalid for use. Skipping inclusion.\n')
                #     continue
                postage_stamp_filename = "{}_{:11.5f}_{:09.5f}_{:+09.5f}.fits".format(obj.provisional_name,
                                                                                      obs.date.mjd,
                                                                                      obs.coordinate.ra.degree,
                                                                                      obs.coordinate.dec.degree)
                print("{}".format(postage_stamp_filename))

                with open(postage_stamp_filename, 'w') as tmp_file:
                    hdulist.writeto(tmp_file, clobber=True)
                    storage.copy(postage_stamp_filename, obj_dir + "/" + postage_stamp_filename)
                os.unlink(postage_stamp_filename)  # easier not to have them hanging around
            except OSError, e:  # occasionally the node is not found: report and move on for later cleanup
                print e
                continue


def main():
    parser = argparse.ArgumentParser(
        description='Parse a directory of TNO .ast files and create links in the postage stamp directory '
                    'that allow retrieval of cutouts of the FITS images associated with the OSSOS detections. '
                    'Cutouts are defined on the WCS RA/DEC of the object position.')

    parser.add_argument("version",
                        help="The OSSOS data release version these stamps should be assigned to.")
    parser.add_argument("--ossin",
                        action="store",
                        default="vos:OSSOS/dbaseclone/ast/",
                        help="The vospace containerNode that clones ossin dbaseclone"
                             "holding the .ast files of astrometry/photometry measurements.")
    parser.add_argument("--blocks", "-b",
                        action="store",
                        default=["o4h"],
                        choices=["o3e", "o3o", "Col3N", 'o3l', 'o4h'],
                        help="Prefixes of object designations to include.")
    parser.add_argument("--radius", '-r',
                        # FIXME: figure out how to assign this a unit.degree before storage
                        action='store',
                        default=0.02 * units.degree,
                        help='Radius (degree) of circle of cutout postage stamp.')
    parser.add_argument("--debug", "-d",
                        action="store_true")
    parser.add_argument("--verbose", "-v",
                        action="store_true")

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)

    for fn in storage.listdir(args.ossin):
        for block in args.blocks:
            if fn.startswith(block):
                obj = mpc.MPCReader(args.ossin + fn)
                obj_dir = '{}/{}/{}'.format(storage.POSTAGE_STAMPS, args.version, obj.provisional_name)
                if not storage.exists(obj_dir, force=True):
                    storage.mkdir(obj_dir)
                # assert storage.exists(obj_dir, force=True)
                print('{} beginning...\n'.format(obj.provisional_name))
                # if int(obj.provisional_name[3:]) == 49:
                assert isinstance(args.radius, Quantity)
                cutout(obj, obj_dir, args.radius)
                print('{} complete.\n'.format(obj.provisional_name))


if __name__ == '__main__':
    main()



__author__ = 'jjk, mtb55'

import os
import re
import struct
import logging

from astropy.coordinates import SkyCoord
from astropy import units
import numpy

from .gui import logger
from .storage import open_vos_or_local
from .util import Time

DEFAULT_OBSERVERS = ['M. T. Bannister',
                     'J. J. Kavelaars',
                     'B. J. Gladman',
                     'J.-M. Petit',
                     'T. Burdullis'
                     ]
DEFAULT_MEASURERS = ['S. D. J. Gwyn',
                     'Y.-T. Chen.'
                     ]
DEFAULT_TELESCOPE = "CFHT 3.6m + CCD"
DEFAULT_ASTROMETRIC_NETWORK = "UCAC4"

MPCNOTES = {"Note1": {" ": " ",
                      "": " ",
                      "*": "*",
                      "A": "earlier approximate position inferior",
                      "a": "sense of motion ambiguous",
                      "B": "bright sky/black or dark plate",
                      "b": "bad seeing",
                      "c": "crowded star field",
                      "D": "declination uncertain",
                      "d": "diffuse image",
                      "E": "at or near edge of plate",
                      "F": "faint image",
                      "f": "involved with emulsion or plate flaw",
                      "G": "poor guiding",
                      "g": "no guiding",
                      "H": "hand measurement of CCD image",
                      "I": "involved with star",
                      "i": "inkdot measured",
                      "J": "J2000.0 rereduction of previously-reported position",
                      "K": "stacked image",
                      "k": "stare-mode observation by scanning system",
                      "M": "measurement difficult",
                      "m": "image tracked on object motion",
                      "N": "near edge of plate, measurement uncertain",
                      "O": "image out of focus",
                      "o": "plate measured in one direction only",
                      "P": "position uncertain",
                      "p": "poor image",
                      "R": "right ascension uncertain",
                      "r": "poor distribution of reference stars",
                      "S": "poor sky",
                      "s": "streaked image",
                      "T": "time uncertain",
                      "t": "trailed image",
                      "U": "uncertain image",
                      "u": "unconfirmed image",
                      "V": "very faint image",
                      "W": "weak image",
                      "w": "weak solution"},
            "Note2": {" ": " ",
                      "": " ",
                      "P": "Photographic",
                      "e": "Encoder",
                      "C": "CCD",
                      "T": "Meridian or transit circle",
                      "M": "Micrometer",
                      "V": "'Roving Observer' observation",
                      "R": "Radar observation",
                      "S": "Satellite observation",
                      "c": "Corrected-without-republication CCD observation",
                      "E": "Occultation-derived observations",
                      "O": "Offset observations (used only for observations of natural satellites)",
                      "H": "Hipparcos geocentric observations",
                      "N": "Normal place",
                      "n": "Mini-normal place derived from averaging observations from video frames"},
            'PhotometryNote': {" ": " ",
                               "": " ",
                               "L": "Photometry uncertainty lacking",
                               "Y": "Photometry measured successfully",
                               "Z": "Photometry measurement failed."}}


class MinorPlanetNumber(object):

    PACKED_DESIGNATION = " ABCDEFGHJKLMNOPQRSTUVWXYZ"

    def __init__(self, minor_planet_number):
        self._minor_planet_number = None
        if minor_planet_number is not None and len(minor_planet_number.strip(' ')) != 0:
            if len(minor_planet_number) == 5:
                if not minor_planet_number[0].isdigit():
                    base = 100000 * (MinorPlanetNumber.PACKED_DESIGNATION.index(minor_planet_number[0]))
                else:
                    base = 10000 * int(minor_planet_number[0])
                self._minor_planet_number = int(minor_planet_number[1:]) + base
            else:
                raise MPCFieldFormatError("minor_planet_number",
                                          "1 char or digit and 4 digits",
                                          minor_planet_number)

    def __str__(self):
        if self._minor_planet_number is None:
            return ""
        if self._minor_planet_number > 99999:
            idx = int(self._minor_planet_number / 100000)
            minor_planet_number = MinorPlanetNumber.PACKED_DESIGNATION[idx]
            minor_planet_number += "{:04d}".format(self._minor_planet_number - idx * 100000)
        else:
            minor_planet_number = "{:05d}".format(self._minor_planet_number)
        return minor_planet_number

    def __int__(self):
        return self._minor_planet_number

    def __cmp__(self, other):
        return cmp(int(self), int(other))

    def __bool__(self):
        return self._minor_planet_number is not None

    def __nonzero__(self):
        return self.__bool__()


class NullObservation(object):

    NULL_OBSERVATION_CHARACTERS = ["!", "-", "#"]

    def __init__(self, null_observation=None, null_observation_character=None):
        """
        A boolean object that keeps track of True/False status via a set of magic characters.
        """
        if null_observation is not None and \
                isinstance(null_observation, basestring) and \
                        len(str(null_observation).strip(' ')) > 0 and \
                null_observation not in NullObservation.NULL_OBSERVATION_CHARACTERS:
            raise MPCFieldFormatError("null_observation",
                                      "one of " + str(NullObservation.NULL_OBSERVATION_CHARACTERS),
                                      null_observation)
        if null_observation_character is None:
            null_observation_character = NullObservation.NULL_OBSERVATION_CHARACTERS[0]
        self.null_observation_character = null_observation_character

        self._null_observation = None
        if isinstance(null_observation, basestring):
            self._null_observation = null_observation in NullObservation.NULL_OBSERVATION_CHARACTERS
        elif isinstance(null_observation, bool):
            self._null_observation = null_observation
        else:
            self._null_observation = False

    def __str__(self):
        return self._null_observation and self.null_observation_character or " "

    def __bool__(self):
        return self._null_observation

    def __nonzero__(self):
        return self.__bool__()


class MPCFormatError(Exception):
    """Base class for errors in MPC formatting."""


class TNOdbFlags(object):
    """
    The OSSOS/CFEPS database has a 'flag' field that indicates OSSOS specific issues associated with an
    MPC formatted line in the database.
    """

    def __init__(self, flags):

        if not re.match("[10]{12}", flags):
            raise ValueError("illegal flag string: {}".format(flags))
        self.__flags = flags

    def __str__(self):
        return self.__flags

    @property
    def is_discovery(self):
        """
        Is this observation part of the discovery triplet?  bit 1
        :return: bool
        """
        return self.__flags[0] == 1

    @is_discovery.setter
    def is_discovery(self, is_discovery):
        self.__flags[0] == bool(is_discovery) and "1" or "0"

    @property
    def is_secret(self):
        """
        Is this observation secret? bit 2
        :return: bool
        """
        return self.__flags[1] == 1


class MPCFieldFormatError(MPCFormatError):
    def __init__(self, field, requirement, actual):
        super(MPCFieldFormatError, self).__init__(
            "Field %s: %s; but was %s" % (field, requirement, actual))


def format_ra_dec(ra_deg, dec_deg):
    """
    Converts RA and DEC values from degrees into the formatting required
    by the Minor Planet Center:

    Formats:
      RA: 'HH MM SS.ddd'
      DEC: 'sDD MM SS.dd' (with 's' being the sign)

    (From: http://www.minorplanetcenter.net/iau/info/OpticalObs.html)

    Args:
      ra_deg: float
        Right ascension in degrees
      dec_deg: float
        Declination in degrees

    Returns:
      formatted_ra: str
      formatted_dec: str
    """
    coords = SkyCoord(ra=ra_deg, dec=dec_deg,
                      unit=(units.degree, units.degree))

    # decimal=False results in using sexagesimal form
    formatted_ra = coords.ra.to_string(unit=units.hour, decimal=False,
                                       sep=" ", precision=3, alwayssign=False,
                                       pad=True)

    formatted_dec = coords.dec.to_string(unit=units.degree, decimal=False,
                                         sep=" ", precision=2, alwayssign=True,
                                         pad=True)

    return formatted_ra, formatted_dec


class MPCNote(object):
    """
    Alphabetic note shown with some of the observations. Non-alphabetic codes are used to differentiate between
    different programs at the same site and such codes will be defined in the headings for the individual
    observatories in the Minor Planet Circulars.
    """

    def __init__(self, code="C", note_type="Note2"):
        self._note_type = None
        self._code = None
        self.note_type = note_type
        self.code = code

    @property
    def note_type(self):
        """
        Note 1 or 2 from an MPC line.
        """
        return self._note_type

    @note_type.setter
    def note_type(self, note_type):
        if note_type not in MPCNOTES.keys():
            raise ValueError("Invalid note_type: expected one of %s got %s" % (str(MPCNOTES.keys()), note_type))
        self._note_type = note_type

    @property
    def code(self):
        """
        The MPC note the denotes the type of detector system used
        """
        return self._code

    @code.setter
    def code(self, code):
        """

        :type code: str
        :param code: an MPC Note code. Either from the allow dictionary or 0-9
        """
        if code is None:
            _code = " "
        else:
            _code = str(code).strip()

        if _code.isdigit():
            if self.note_type != 'Note1':
                raise MPCFieldFormatError(self.note_type,
                                          "Must be a character",
                                          _code)
            if not 0 < int(_code) < 10:
                raise MPCFieldFormatError(self.note_type,
                                          "numeric value must be between 0 and 9",
                                          _code)
        else:
            if len(_code) > 1:
                raise MPCFieldFormatError(self.note_type,
                                          "must be 0 or 1 characters",
                                          _code)
            if _code not in MPCNOTES[self.note_type]:
                raise MPCFieldFormatError(self.note_type,
                                          "must one of " + str(MPCNOTES[self.note_type]),
                                          _code)
        self._code = _code

    def __str__(self):
        return str(self.code)

    @property
    def long(self):
        return MPCNOTES[self.note_type][self.code]


class Discovery(object):
    """
    Holds the discovery flag for an MPC Observation Line
    """

    def __init__(self, is_discovery=False):
        self._is_discovery = False
        self._is_initial_discovery = False
        self.is_discovery = is_discovery
        self.is_initial_discovery = is_discovery

    def set_from_mpc_line(self, mpc_line):
        """
        Given an MPC line set the discovery object
        """
        mpc_line = str(mpc_line)
        if len(mpc_line) < 56:
            raise MPCFieldFormatError("mpc_line",
                                      "is too short",
                                      mpc_line)
        self.is_discovery = mpc_line[12]

    @property
    def is_initial_discovery(self):
        return self._is_initial_discovery

    @property
    def is_discovery(self):
        return self._is_discovery

    @is_discovery.setter
    def is_discovery(self, is_discovery):
        if is_discovery not in ['*', '&', ' ', '', True, False, None]:
            raise MPCFieldFormatError("discovery",
                                      "must be one of '',' ','&', '*',True, False. ",
                                      is_discovery)
        self._is_discovery = is_discovery in ['*', '&', True] and True or False

    @is_initial_discovery.setter
    def is_initial_discovery(self, is_discovery):
        """
        Is this MPC line the initial discovery line?
        @param is_discovery: the code for the discovery setting "*" or True or False
        """
        self._is_initial_discovery = (is_discovery in ["*", True] and True) or False

    def __str__(self):
        if self.is_initial_discovery:
            return "*"
        if self.is_discovery:
            return "&"
        return " "

    def __bool__(self):
        return self.is_discovery

    def __nonzero__(self):
        return self.__bool__()


def compute_precision(coord):
    """
    Returns the number of digits after the last '.' in a given number or string.

    """
    coord = str(coord).strip(' ')
    idx = coord.rfind('.')
    if idx < 0:
        return 0
    else:
        return len(coord) - idx - 1


def get_date(date_string):
    """
    Given an MPC formatted time string return a Time object.

    :rtype : Time
    :param date_string: a string in MPC date format
    """
    _date_precision = compute_precision(date_string)
    return Time(date_string, format='mpc', scale='utc', precision=_date_precision)


class Observation(object):
    """
    An observation of an object, nominally generated by reading an MPC formatted file.
    """

    def __init__(self,
                 null_observation=False,
                 minor_planet_number=None,
                 provisional_name=None,
                 discovery=False,
                 note1=None,
                 note2=None,
                 date="2000 01 01.000001",
                 ra="00 00 00.000",
                 dec="+00 00 00.00",
                 mag=-1,
                 band='r',
                 observatory_code=568,
                 comment=None,
                 mag_err=-1,
                 xpos=None,
                 ypos=None,
                 frame=None,
                 plate_uncertainty=None,
                 astrometric_level=0):
        """

        :param provisional_name:
        :param discovery:
        :param note1:
        :param note2:
        :param date:
        :param ra:
        :param dec:
        :param mag:
        :param band:
        :param observatory_code:
        :param comment: A comment about this observation, not sent
        :param mag_err:
        :param xpos:
        :param ypos:
        :param frame:
        :param plate_uncertainty:
        :param null_observation:

        :type comment MPCComment
        """
        self._null_observation = False
        self._minor_planet_number = None
        self._provisional_name = None

        self.null_observation = null_observation
        self.minor_planet_number = minor_planet_number
        self.provisional_name = provisional_name

        self._discovery = None
        self.discovery = discovery
        self._note1 = None
        self.note1 = note1
        self._note2 = None
        self.note2 = note2
        self._date = None
        self._date_precision = None
        self.date = date
        self._coordinate = None
        self.coordinate = (ra, dec)
        self._mag = None
        self._mag_err = None
        self._mag_precision = 1
        self._ra_precision = 3
        self._dec_precision = 2
        self.mag = mag
        self.mag_err = mag_err
        self._band = None
        self.band = band
        self._observatory_code = None
        self.observatory_code = observatory_code
        self._comment = None
        self.comment = OSSOSComment(version="O", frame=frame,
                                    source_name=provisional_name,
                                    photometry_note="",
                                    mpc_note=str(self.note1),
                                    x=xpos,
                                    y=ypos,
                                    plate_uncertainty=plate_uncertainty,
                                    astrometric_level=astrometric_level,
                                    magnitude=mag,
                                    mag_uncertainty=mag_err,
                                    comment=comment)

    def __eq__(self, other):
        return self.to_string() == other.to_string()

    def __ne__(self, other):
        return self.to_string() != other.to_string()

    def __le__(self, other):
        return self.date <= other.date

    def __lt__(self, other):
        return self.date < other.date

    def __ge__(self, other):
        return self.date >= other.date

    def __gt__(self, other):
        return self.date > other.date

    @classmethod
    def from_ted(cls, ted):
        """
        Turn a ted line into an MPC formatted line. (Marc Buie's format)

        Example line: 2011 04  28.29389  18 32 37.260  -21 13 49.86  26.3R NI100     304
        :param ted: a TED formatted minor planet observations
        :return: Observation
        """
        ted = ted.strip()
        if len(ted) != len('2011 04  28.29389  18 32 37.260  -21 13 49.86  26.3R NI100     304'):
            raise ValueError("Incorrectly formatted line for ted to mpc conversion.")
        line_order = ["date", "ra", "dec", "mag", "filter", "provisional_name", "observatory_code"]
        parts = {"date": "2011 04  28.29389  ",
                 "ra": "18 32 37.260  ",
                 "dec": "-21 13 49.86  ",
                 "mag": "26.3",
                 "filter": "R ",
                 "provisional_name": "NI100     ",
                 "observatory_code": "304"}
        end_pos = 0
        args = {}
        for part in line_order:
            start_pos = end_pos
            end_pos = start_pos + len(parts[part])
            args[part] = ted[start_pos:end_pos]
        return Observation(provisional_name=args['provisional_name'],
                           date=args['date'],
                           ra=args['ra'],
                           dec=args['dec'],
                           mag=args['mag'],
                           band=args['filter'],
                           observatory_code=args['observatory_code'])

    @classmethod
    def from_string(cls, input_line):
        """
        Given an MPC formatted line, returns an MPC Observation object.

        This method attempts four different format.

        If line is less than 80 characters then assumes in 'ted' format,

        If line equal/greater than 80 tries some fixed structure formats

        mpc_format = '0s5s7s1s1s1s17s12s12s9x5s1s6x3s'
        ossos_format1 = '1s4s7s1s1s1s17s12s12s9x5s1s6x3s'
        ossos_format2' = '1s0s11s1s1s1s17s12s12s9x5s1s6x3s'

        if those fail then tries Alex Parker's .ast format.

        :param input_line: a line in the one-line roving observer format
        :type input_line: str
        """
        struct_formats = {'mpc_format': '0s5s7s1s1s1s17s12s12s9x5s1s6x3s',
                          'ossos_format1': '1s4s7s1s1s1s17s12s12s9x5s1s6x3s',
                          'ossos_format2': '1s0s11s1s1s1s17s12s12s9x5s1s6x3s'}

        mpc_line = input_line.strip('\n')
        logging.debug("Trying to create MPC record from:\n{}".format(mpc_line))
        if len(mpc_line) > 0 and mpc_line[0] == '#':
            return MPCComment.from_string(mpc_line[1:])

        comment = mpc_line[81:]
        mpc_line = mpc_line[0:80]
        if len(mpc_line) != 80:
            logging.error("Line is short, trying .ted format")
            try:
                return cls.from_ted(mpc_line)
            except Exception as ex:
                logging.debug(type(ex))
                logging.debug(str(ex))

        obsrec = None
        for format_name in struct_formats:
            try:
                logging.debug("Trying to parse with {}".format(format_name))
                obsrec = cls(*struct.unpack(struct_formats[format_name], mpc_line))
                break
            except Exception as ex:
                logging.debug("Failed: {}".format(ex))
                obsrec = None

        if obsrec is None:
            logging.debug("Trying AP's .ast format")
            # try converting using Alex Parker's .ast format:
            # 2456477.78468 18:39:07.298 -20:40:17.53 0.2 304
            try:
                _parts = mpc_line.split(' ')
                args = {"date": Time(float(_parts[0]), scale='utc', format='jd').mpc,
                        "discovery": False,
                        "ra": _parts[1].replace(":", " "),
                        "dec": _parts[2].replace(":", " "),
                        "plate_uncertainty": _parts[3],
                        "observatory_code": _parts[4]}
                obsrec = cls(**args)
            except:
                obsrec = None

        if not obsrec:
            logging.error("Failed to parse line: {}".format(mpc_line))
            return obsrec

        obsrec.comment = MPCComment.from_string(comment)
        if isinstance(obsrec.comment, OSSOSComment) and obsrec.comment.source_name is None:
            obsrec.comment.source_name = obsrec.provisional_name
        # Check if there are TNOdb style flag lines.
        if isinstance(obsrec.comment, TNOdbComment):
            if obsrec.comment.flags[0] == '1':
                obsrec.discovery.is_discovery = True

        return obsrec

    def to_string(self):
        as_string = str(self)
        if self.comment is not None and str(self.comment) != "":
            as_string += " " + str(self.comment)
        return as_string

    def __str__(self):
        """
        Writes out data about accepted objects in the Minor Planet Center's 'Minor Planets'
        format as specified here:
        http://www.minorplanetcenter.net/iau/info/OpticalObs.html
        """
        # MOP/OSSOS allows the provisional name to take up the full space allocated to the MinorPlanetNumber AND
        # the provisional name.

        if not self.minor_planet_number:
            if len(self.provisional_name) > 7:
                mpc_str = "{:1.1s}{:<11.11s}".format(self.null_observation,
                                                     self.provisional_name)
            else:
                mpc_str = "{:1.1s}{:4.4s}{:<7.7s}".format(self.null_observation,
                                                          " " * 4,
                                                          self.provisional_name)
        else:
            mpc_str = "{:5.5s}{:<7.7s}".format(self.minor_planet_number, self.provisional_name)

        mpc_str += str(self.discovery)
        mpc_str += '{0:1s}{1:1s}'.format(str(self.note1), str(self.note2))
        mpc_str += '{0:<17s}'.format(str(self.date))
        mpc_str += '{0:<12s}{1:<12s}'.format(str(self.ra), str(self.dec))
        mpc_str += 9 * " "
        mag_format = '{0:<5.' + str(self._mag_precision) + 'f}{1:1s}'
        band = self.band is None and " " or self.band
        mag_str = (self.mag is None and 6 * " ") or mag_format.format(self.mag, band)
        if len(mag_str) != 6:
            raise MPCFieldFormatError("mag",
                                      "length of mag string should be exactly 6 characters, got->",
                                      mag_str)
        mpc_str += mag_str
        mpc_str += 6 * " "
        mpc_str += "%3s" % self.observatory_code
        return mpc_str

    def to_tnodb(self):
        """
        provide string representation of observation in a format used for OSSOS database input.
        """

        # O indicates OSSOS survey
        if not isinstance(self.comment, OSSOSComment):
            logging.warn("Non OSSOS comment:{}".format(self.comment))

        comment_line = "#"+str(self.comment).rstrip('\n')

        if self.mag == -1:  # write no mag and no filter for where photometry couldn't be measured
            self.mag = None
        else:
            # set mag precision back to 0.1 mags regardless of how good it actually is
            self._mag_precision = 1

        # set the null observation character to the tnodb value
        self.null_observation.null_observation_character = "-"
        mpc_observation = str(self)

        return comment_line + '\n' + mpc_observation

    def to_mpc(self):
        self.null_observation.null_observation_character = "#"
        return str(self)

    @property
    def null_observation(self):
        return self._null_observation

    @null_observation.setter
    def null_observation(self, null_observation=False):
        """
        :param null_observation: is this a null observation marker True/False
        """
        self._null_observation = NullObservation(null_observation)

    @property
    def provisional_name(self):
        return self._provisional_name

    @provisional_name.setter
    def provisional_name(self, provisional_name=None):
        if provisional_name is None:
            provisional_name = " " * 7
        else:
            provisional_name = provisional_name.strip()
            # if not provisional_name[0].isalpha():
            # logging.warning("Provisional Name should not be a number: {}".format(provisional_name))
            # if not len(provisional_name) <= 7:
            #     logging.warning("Provisional Name too long {}".format(provisional_name))
        self._provisional_name = provisional_name

    @property
    def minor_planet_number(self):
        """
        :return: minor_planet_number for object associated with this observation.
        """
        return self._minor_planet_number

    @minor_planet_number.setter
    def minor_planet_number(self, minor_planet_number):
        self._minor_planet_number = MinorPlanetNumber(minor_planet_number)

    @property
    def discovery(self):
        """
        Is this a discovery observation?

        :return True/False
        """
        return self._discovery

    @discovery.setter
    def discovery(self, is_discovery):
        """

        :type is_discovery: bool
        :param is_discovery: indicates if observation was a discovery
        """
        self._discovery = Discovery(is_discovery=is_discovery)

    @property
    def note1(self):
        return self._note1

    @note1.setter
    def note1(self, note1):
        self._note1 = MPCNote(code=note1, note_type="Note1")

    @property
    def note2(self):
        return self._note2

    @note2.setter
    def note2(self, code):
        self._note2 = MPCNote(code=code, note_type="Note2")

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, date_str):
        self._date_precision = compute_precision(date_str)
        logging.debug("Setting precision to: {}".format(self._date_precision))
        try:
            self._date = Time(date_str, format='mpc', scale='utc', precision=self._date_precision)
        except Exception as ex:
            logging.error(str(ex))
            raise MPCFieldFormatError("Observation Date",
                                      "does not match expected format",
                                      date_str)

    @property
    def ra(self):
        return self.coordinate.ra.to_string(unit=units.hour, decimal=False,
                                            sep=" ", precision=self._ra_precision, alwayssign=False,
                                            pad=True)

    @property
    def dec(self):
        return self.coordinate.dec.to_string(unit=units.degree, decimal=False,
                                             sep=" ", precision=self._dec_precision, alwayssign=True,
                                             pad=True)

    @property
    def comment(self):
        """

        :return: the comment
        :rtype: OSSOSComment
        """
        return self._comment

    @comment.setter
    def comment(self, comment):
        if comment is None:
            self._comment = ""
        else:
            self._comment = comment

    @property
    def coordinate(self):
        return self._coordinate

    @coordinate.setter
    def coordinate(self, coord_pair):
        """

        :param coord_pair: RA/DEC pair [as a tuple or single string]
        """

        if type(coord_pair) in [list, tuple] and len(coord_pair) == 2:
            val1 = coord_pair[0]
            val2 = coord_pair[1]
        else:
            raise MPCFieldFormatError("RA/DEC",
                                      "Expected a pair of coordinates got: ",
                                      coord_pair)

        self._ra_precision = 3
        self._dec_precision = 2
        try:
            ra = float(val1)
            dec = float(val2)
            self._coordinate = SkyCoord(ra, dec, unit=(units.degree, units.degree))
        except:
            try:
                self._ra_precision = compute_precision(val1)
                self._ra_precision = self._ra_precision < 3 and self._ra_precision or 3
                self._dec_precision = compute_precision(val2)
                self._dec_precision = self._dec_precision < 2 and self._dec_precision or 2
                self._coordinate = SkyCoord(val1, val2, unit=(units.hour, units.degree))
            except Exception as ex:
                logging.debug(type(ex))
                logging.debug(str(ex))
                raise MPCFieldFormatError("coord_pair",
                                          "must be [ra_deg, dec_deg] or HH MM SS.S[+-]dd mm ss.ss",
                                          coord_pair)

    @property
    def mag(self):
        return self._mag

    @mag.setter
    def mag(self, mag):
        if mag is None or len(str(str(mag).strip(' '))) == 0 or float(mag) < 0:
            self._mag_precision = 0
            self._mag = None
        else:
            self._mag = float(mag)
            self._mag_precision = min(1, compute_precision(str(mag)))

    @property
    def mag_err(self):
        return self._mag_err

    @mag_err.setter
    def mag_err(self, mag_err):
        if mag_err is None or len(str(mag_err).strip('')) == 0 or self.mag is None:
            self._mag_err = None
        else:
            self._mag_err = mag_err

    @property
    def band(self):
        return self._band

    @band.setter
    def band(self, band):
        band = str(band.strip(' '))
        self._band = (len(band) > 0 and str(band)[0]) or None

    @property
    def observatory_code(self):
        return self._observatory_code

    @observatory_code.setter
    def observatory_code(self, observatory_code):
        observatory_code = str(observatory_code)
        if not len(observatory_code) <= 3:
            raise MPCFieldFormatError("Observatory code",
                                      "must be 3 characters or less",
                                      observatory_code)
        self._observatory_code = str(observatory_code)


class OSSOSComment(object):
    """
    Parses an OSSOS observation's metadata into a format that can be stored in the 
    an Observation.comment and written out in the same MPC line.

    Specification: '1s1x12s1x11s1x2s1x7s1x7s1x4s1x1s1x5s1x4s1x'
    """

    def __init__(self, version, frame, source_name, photometry_note, mpc_note, x, y,
                 plate_uncertainty=0.2,
                 astrometric_level=0,
                 magnitude=None,
                 mag_uncertainty=None,
                 comment=None):

        self.version = version
        self.frame = frame
        self.source_name = source_name
        self._photometry_note = None
        self.photometry_note = photometry_note
        self.mpc_note = mpc_note
        self._x = None
        self.x = x
        self._y = None
        self.y = y
        self._mag = None
        self.mag = magnitude
        self._mag_uncertainty = None
        self.mag_uncertainty = mag_uncertainty
        self._plate_uncertainty = None
        self.plate_uncertainty = plate_uncertainty
        self._astrometric_level = 0
        self.astrometric_level = astrometric_level
        self._comment = ""
        self.comment = comment
        self.flags = None

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def __le__(self, other):
        raise NotImplemented

    def __ge__(self, other):
        raise NotImplemented

    @classmethod
    def from_string(cls, comment):
        """
        Build an MPC Comment from a string.
        """
        if comment is None or len(comment) == 0:
            return str("")
        if comment[0] == "#":
            comment = comment[1:]
        values = comment.split('%')
        comment_string = ""
        if len(values) > 1:
            comment_string = values[1].lstrip(' ')
        # O 1631355p21 O13AE2O     Z  1632.20 1102.70 0.21 3 ----- ---- % Apcor failure.
        ossos_comment_format = '1s1x12s1x11s1x1s1s1x7s1x7s1x4s1x1s1x5s1x4s1x'
        old_ossos_comment_format = '1s1x10s1x11s1x1s1s1x7s1x7s1x4s1x1s1x5s1x4s1x'
        for struct_ in [ossos_comment_format, old_ossos_comment_format]:
            try:
                retval = cls(*struct.unpack(struct_, values[0]))
                retval.comment = values[1]
                return retval
            except Exception as e:
                logging.debug(str(e))
                logging.debug("OSSOS Fixed Format Failed.")
                logging.debug(comment)
                logging.debug("Trying space separated version")

        values = values[0].split()
        try:
            if values[0] != 'O' or len(values) < 6:
                # this is NOT and OSSOS style comment string
                raise ValueError("Can't parse non-OSSOS style comment: {}".format(comment))
            # first build a comment based on the required fields.
            retval = cls(version="O",
                         frame=values[1],
                         source_name=values[2],
                         photometry_note=values[3][0],
                         mpc_note=values[3][1:],
                         x=values[4],
                         y=values[5],
                         comment=comment_string)
        except Exception as e:
            logging.error(values)
            logging.error(comment)
            logging.error(str(e))
            raise e

        retval.version = values[0]
        logging.debug("length of values: {}".format(len(values)))
        logging.debug("Values: {}".format(str(values)))
        # the format of the last section evolved during the survey, but the following flags should handle this.
        if len(values) == 7:
            retval.plate_uncertainty = values[6]
        elif len(values) == 8:
            retval.plate_uncertainty = values[6]
            retval.astrometric_level = values[7]
        elif len(values) == 9:  # This is the old format where mag was in-between X/Y and uncertainty in X/Y
            retval.mag = values[6]
            retval.mag_uncertainty = values[7]
            retval.plate_uncertainty = values[8]
        elif len(values) == 10:  # if there are 9 values then the new astrometric level value is set.
            retval.plate_uncertainty = values[8]
            retval.astrometric_level = values[9]
            logging.debug('here now')
            retval.mag = values[6]
            retval.mag_uncertainty = values[7]
        logging.debug("DONE.")
        return retval

    @property
    def mag(self):
        return self._mag

    @mag.setter
    def mag(self, mag):
        try:
            self._mag = float(mag)
            self.photometry_note = "Y"
            if not 15 < self._mag < 30:
                raise ValueError("Magnitude out of reasonable range:  15 < mag < 30")
        except:
            self.photometry_note = "Z"
            self._mag = None

    @property
    def mag_uncertainty(self):
        return self._mag_uncertainty

    @mag_uncertainty.setter
    def mag_uncertainty(self, mag_uncertainty):
        try:
            self._mag_uncertainty = float(mag_uncertainty)
            if not 0 < self._mag_uncertainty < 1.0:
                raise ValueError("mag uncertainty must be in range 0 to 1")
        except Exception as e:
            logging.debug("Failed trying to convert merr ({}) to float. Using default.".format(mag_uncertainty))
            logging.debug(str(e))
            if str(self.mag).isdigit():
                self.photometry_note = "L"
            else:
                self.photometry_note = "Z"
            self._mag_uncertainty = None

    @property
    def photometry_note(self):
        return self._photometry_note

    @property
    def astrometric_level(self):
        return self._astrometric_level

    @astrometric_level.setter
    def astrometric_level(self, astrometric_level):
        astrometric_level = int(astrometric_level)
        if not -1 < astrometric_level < 10:
            raise ValueError("Astrometric level must be integer between 0 and 9.")
        self._astrometric_level = astrometric_level

    @photometry_note.setter
    def photometry_note(self, photometry_note):
        self._photometry_note = str(photometry_note)

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        try:
            self._x = float(x)
        except:
            self._x = None

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, y):
        try:
            self._y = float(y)
        except:
            self._y = None

    @property
    def plate_uncertainty(self):
        return self._plate_uncertainty

    @plate_uncertainty.setter
    def plate_uncertainty(self, plate_uncertainty):
        try:
            self._plate_uncertainty = float(plate_uncertainty)
        except:
            self._plate_uncertainty = 0.2
        if not 0 < self._plate_uncertainty < 100:
            raise ValueError("Plate uncertainty must be between 0 and 100. (in arc-seconds)")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        if comment is not None:
            try:
                self._comment = str(comment.strip())
            except:
                self._comment = ''
        else:
            self._comment = ''

    @staticmethod
    def to_str(frmt, value, default="", sep=" "):
        try:
            if value is None:
                raise ValueError("Don't print None.")
            return sep+frmt.format(value)
        except:
            return sep+default

    def __str__(self):
        """
        Format comment as required for storing OSSOS metadata
        odonum p ccd object_name MPCnotes X Y mag mag_uncertainty plate_uncertainty % comment
        """
        # The astrometric uncertainty should be set to higher when hand measurements are made.
        if self.version == "T":
            return self.comment

        if self.version == "L":
            return "{:1s} {:10s} {}".format(self.version, self.frame, self.comment)

        comm = '{:1s}'.format(self.version)
        comm += self.to_str("{:>12.12s}", self.frame, "-"*12)
        comm += self.to_str("{:<11.11s}", self.source_name, "-"*11)
        comm += self.to_str("{:2.2s}", self.photometry_note+self.mpc_note, "--")
        comm += self.to_str("{:>7.2f}", self.x, "-"*7)
        comm += self.to_str("{:>7.2f}", self.y, "-"*7)
        comm += self.to_str('{:4.2f}', self.plate_uncertainty, "-"*4)
        comm += self.to_str('{:1d}', self.astrometric_level, "-")
        comm += self.to_str('{:5.2f}', self.mag, "-"*5)
        comm += self.to_str('{:4.2f}', self.mag_uncertainty, "-"*4)
        comm += ' % {}'.format(self.comment)

        return comm


class MPCWriter(object):
    """
    Writes out data about accepted objects in the Minor Planet Center's
    format as specified here:
    http://www.minorplanetcenter.net/iau/info/OpticalObs.html

    Note that we assume objects fall under the Minor Planet category.

    Format reproduced below for convenience:

        Columns     Format   Use
        1 -  5        A5     Minor planet number
        6 - 12        A7     Provisional or temporary designation
        13            A1     Discovery asterisk
        14            A1     Note 1
        15            A1     Note 2
        16 - 32      A17     Date of observation
        33 - 44      A12     Observed RA (J2000.0)
        45 - 56      A12     Observed Decl. (J2000.0)
        57 - 65       9X     Must be blank
        66 - 71    F5.2,A1   Observed magnitude and band
                               (or nuclear/total flag for comets)
        72 - 77       6X     Must be blank
        78 - 80       A3     Observatory code
    """

    def __init__(self, file_handle, auto_flush=True, include_comments=True,
                 auto_discovery=True, formatter=None):
        self.filehandle = file_handle
        self.auto_flush = auto_flush
        self.include_comments = include_comments

        # Holds observations that have not yet been flushed
        self.buffer = {}
        self._written_mpc_observations = []

        self.auto_discovery = auto_discovery
        self._discovery_written = False
        if formatter is None:
            if self.include_comments:
                self.formatter = Observation.to_string
            else:
                self.formatter = Observation.__str__
        else:
            self.formatter = formatter

    def get_filename(self):
        return self.filehandle.name

    def write(self, mpc_observation):
        """
        Writes a single entry in the Minor Planet Center's format.
        :param mpc_observation:
        """
        assert isinstance(mpc_observation, Observation)
        try:
            key = mpc_observation.comment.frame.strip()
        except:
            key = mpc_observation.date.mjd

        try:
            # keep any 'discovery' flags that are set on observation if already in buffer.
            mpc_observation.discovery = self.buffer[key].discovery.is_discovery
        except Exception as ex:
            logger.debug(type(ex))
            logger.debug(str(ex))
            pass
        self.buffer[key] = mpc_observation

        if self.auto_flush:
            self.flush()

    def flush(self):
        for obs in self.get_chronological_buffered_observations():
            self._flush_observation(obs)

        self.filehandle.flush()

    def _flush_observation(self, obs):
        assert isinstance(obs, Observation)
        if (self.auto_discovery and
                not obs.null_observation and
                not self._discovery_written):
            obs.discovery = True

        if obs.discovery:
            if self._discovery_written:
                obs.discovery.is_initial_discovery = False
            else:
                self._discovery_written = True

        if obs.date.jd not in self._written_mpc_observations:
            self._written_mpc_observations.append(obs.date.jd)
            line = self.formatter(obs)
            self.filehandle.write(line + "\n")

    def close(self):
        self.filehandle.close()

    def get_chronological_buffered_observations(self):
        jds = self.buffer.keys()
        jds.sort()
        sorted_obs = []
        for jd in jds:
            sorted_obs.append(self.buffer[jd])
        return sorted_obs


def make_tnodb_header(observations,
                      observatory_code=None,
                      observers=DEFAULT_OBSERVERS,
                      measurers=DEFAULT_MEASURERS,
                      telescope=DEFAULT_TELESCOPE,
                      astrometric_network=DEFAULT_ASTROMETRIC_NETWORK):
    """
    Write a header appropriate for a tnodb style of file.
    """
    observatory_code = observatory_code is None and observations[0].observatory_code or observatory_code

    odates = [obs.date for obs in observations]
    mindate = min(odates).iso.replace('-', '')[0:8]
    maxdate = max(odates).iso.replace('-', '')[0:8]

    header = "COD {}\n".format(observatory_code)

    sep = ""
    header += "OBS "
    for observer in observers[:-1]:
        header += "{}{}".format(sep, observer)
        sep = ", "
    if len(observers) > 1:
        header += " and {}".format(observers[-1])
    header += "\n"
    header += "MEA "
    sep = ""
    for measurer in measurers[:-1]:
        header += "{}{}".format(sep, measurer)
        sep = ", "
    if len(measurers) > 1:
        header += " and {}".format(measurers[-1])
    header += "\n"
    header += "TEL {}\n".format(telescope)
    header += "NET {}\n".format(astrometric_network)
    header += "{:s} {:s}\n".format('STD', mindate)
    header += "{:s} {:s}\n".format('END', maxdate)

    return header


class MPCReader(object):
    """
    A class to read in MPC files.

    Can be initialized with a filename, will then initialize the mpc_observations attribute to hold the observations.
    """

    def __init__(self, filename=None, replace_provisional=None, provisional_name=None):
        self.replace_provisional = replace_provisional
        self._provisional_name = provisional_name
        if filename is not None:
            self.filename = filename
            self.mpc_observations = self.read(filename)

    def read(self, filename):
        """
        Read  MPC records from filename:

        :param filename: filename of file like object.
        :rtype : [Observation]
        """

        self.filename = filename
        # can be a file like objects,
        if isinstance(filename, basestring):
            filehandle = open_vos_or_local(filename, "rb")
        else:
            filehandle = filename

        input_mpc_lines = filehandle.read().split('\n')
        filehandle.close()
        mpc_observations = []
        next_comment = None
        for line in input_mpc_lines:
            line = line.rstrip()
            mpc_observation = Observation.from_string(line)
            if isinstance(mpc_observation, OSSOSComment):
                next_comment = mpc_observation
                continue
            if isinstance(mpc_observation, Observation):
                if next_comment is not None:
                    mpc_observation.comment = next_comment
                    next_comment = None

                if self.replace_provisional is not None:  # then it has an OSSOS designation: set that in preference
                    mpc_observation.provisional_name = self.provisional_name
                if len(str(mpc_observation.provisional_name.strip())) == 0 and \
                                str(mpc_observation.minor_planet_number) == "":
                    mpc_observation.provisional_name = filename.split(".")[0]
                mpc_observations.append(mpc_observation)

        # No assurance that a .ast file is date-ordered: date-ordered is more expected behaviour
        mpc_observations.sort(key=lambda x: x.date)

        return numpy.array(mpc_observations)

    @property
    def provisional_name(self):
        """
        Determine the provisional name based on the file being accessed.
        :return: str
        """
        if self._provisional_name is not None:
            return self._provisional_name
        if isinstance(self.filename, basestring):
            self._provisional_name = self.filename.rstrip('.ast')
        elif hasattr(self.filename, 'name'):
            self._provisional_name = self.filename.name
        elif hasattr(self.filename, 'filename'):
            self._provisional_name = self.filename.filename
        elif hasattr(self.filename, '__class__'):
            self._provisional_name = str(self.filename.__class__)
        else:
            self._provisional_name = str(type(self.filename))
        self._provisional_name = os.path.basename(self._provisional_name)
        return self._provisional_name


class Index(object):
    """
    MOP/OSSOS name mapping index.
    """
    MAX_NAME_LENGTH = 10

    def __init__(self, idx_filename):
        self.names = {}
        self.index = {}
        with open(idx_filename, 'r') as idx_handle:
            for line in idx_handle.readlines():
                master_name = line[0:Index.MAX_NAME_LENGTH]
                master_name = master_name.strip()
                self.names[master_name] = master_name
                self.index[master_name] = [master_name]
                for i in range(Index.MAX_NAME_LENGTH, len(line), Index.MAX_NAME_LENGTH):
                    this_name = line[i:i + Index.MAX_NAME_LENGTH].strip()
                    self.index[master_name].append(this_name)
                    self.names[this_name] = master_name

    def __str__(self):
        result = ""
        for name in self.index:
            result += "{0:<{1}s}".format(name, Index.MAX_NAME_LENGTH)
            for alias in self.get_aliases(name):
                result += "{0:<{1}s}".format(alias, Index.MAX_NAME_LENGTH)
            result += "\n"
        return result

    def get_aliases(self, name):
        """
        get all names associated with a given name.
        :rtype : list
        :param name: object to get alias names of.
        """
        if name not in self.names:
            return name
        return self.index[self.names[name]]

    def is_same(self, name1, name2):
        """
        Do name1 and name2 refer to the same object?

        :param name1: name of object 1
        :param name2: name of object 2
        :return: Bool
        """
        return name2 in self.get_aliases(name1)


class MPCConverter(object):
    """
    Converts an MPC formatted file to a TNOdb one.
    :param mpc_file The input filename, of MPC lines.
    :param output   if required; else will use root of provided MPC file.

    batch_convert is factory method that will write out a series of input files given an input path.
    """

    def __init__(self, mpc_file, output=None):

        if output is None:
            output = mpc_file.rpartition('.')[0] + '.tnodb'

        self.mpc_file = mpc_file
        self.outfile = open(output, 'w')
        self.write_header = True

    def convert(self):
        with open(self.mpc_file, 'r') as infile:
            observations = []
            for line in infile.readlines():
                obs = Observation().from_string(line)
                observations.append(obs)

            if self.write_header:
                self.outfile.write(make_tnodb_header(observations))
                self.write_header = False

            for obs in observations:
                self.outfile.write(obs.to_tnodb() + '\n')

    @classmethod
    def batch_convert(cls, path):
        for fn in os.listdir(path):
            if fn.endswith('.mpc') or fn.endswith('.track') or fn.endswith('.checkup') or fn.endswith('.nailing'):
                cls(path + fn).convert()


class CFEPSComment(OSSOSComment):
    """
    This holds the old-style comments that come for CFEPS style entries.
    """
    def __init__(self, frame, comment):

        if "measured inside confirm @" in comment:
            values = comment.split('@')[1].split()
            x = values[0]
            y = values[1]
        else:
            x = ""
            y = ""
        source_name = None
        mpc_note = " "
        super(CFEPSComment, self).__init__("O", frame, source_name, " ", mpc_note, x, y, comment=comment)
        self.version = "L"

    @classmethod
    def from_string(cls, comment):
        """
        Build a comment from a CFEPS style comment string.
        """
        values = comment.split()
        if values[0] != "L" or len(values) < 2:
            raise ValueError("Not a CFEPS style comment: {}".format(comment))
        frame = values[1]
        comment = " ".join(values[2:])
        return cls(frame, comment)


class TNOdbComment(OSSOSComment):
    """
    This holds a TNOdb style comment line which contains flags.

    A TNOdb style comment consists of three space seperated fields that are used by the tnodb followed by a
    comment string that is either in the CFEPS or OSSOS format.
    """
    def __init__(self, index, date, flags, **kwargs):

        super(TNOdbComment, self).__init__(**kwargs)
        self.index = index
        self.date = date
        self.flags = flags

    @classmethod
    def from_string(cls, line):
        if len(line) < 56:
            raise ValueError("Line too short, not a valid TNOdb comment string: {}".format(line))
        index = line[0:14].strip()
        if not re.match(r'\d{8}_\S{3}_\S', index):
            raise ValueError("No valid flags, not a valid TNOdb comment string: {}".format(line))

        date = line[15:23].strip()
        flags = line[24:34].strip()
        comment = line[56:].strip()

        comment_object = None
        # try build a comment object based on the TNOdb comment string
        if len(comment) > 0:
            for func in [OSSOSComment.from_string,
                         CFEPSComment.from_string]:
                try:
                    comment_object = func(comment)
                except ValueError as verr:
                    logging.debug(verr)
                    continue
                break

        if isinstance(comment_object, OSSOSComment):
            retval = cls(index,
                         date,
                         flags,
                         version=comment_object.version,
                         frame=comment_object.frame,
                         source_name=comment_object.source_name,
                         mpc_note=comment_object.mpc_note,
                         x=comment_object.x,
                         y=comment_object.y,
                         magnitude=comment_object.mag,
                         mag_uncertainty=comment_object.mag_uncertainty,
                         photometry_note=comment_object.photometry_note,
                         plate_uncertainty=comment_object.plate_uncertainty,
                         astrometric_level=comment_object.astrometric_level,
                         comment=comment_object.comment)
            return retval
        else:
            retval = cls(index,
                         date,
                         flags,
                         version="T",
                         frame=" ",
                         source_name="",
                         photometry_note=" ",
                         mpc_note=" ",
                         x=" ",
                         y=" ",
                         comment=comment)
            return retval

    def to_string(self):
        comm = "{} {} {}".format(self.index, self.date, self.flags)
        # add 22 spaces that are currently padding in TNOdb style records.
        comm += " "*22
        comm += str(self)
        return comm


class RealOSSOSComment(OSSOSComment):

    @classmethod
    def from_string(cls, comment):
        if comment.strip()[0] != "O":
            comment = "O "+comment
        return super(RealOSSOSComment, cls).from_string(comment)


class MPCComment(OSSOSComment):
    """
    A generic class for all comment strings.. try and figure out which one to use.
    """

    @classmethod
    def from_string(cls, line):
        comment = line
        logging.debug('Here is the comment. ' + comment)
        for func in [TNOdbComment.from_string,
                     OSSOSComment.from_string,
                     RealOSSOSComment.from_string,
                     CFEPSComment.from_string,
                     str]:
            try:
                comment = func(line)
            except ValueError as verr:
                logging.debug(str(func))
                logging.debug(str(verr))
                continue
            break
        return comment
